import itertools
from functools import partial
from typing import *

import hypothesis
import hypothesis.strategies as st
import pytest
import torch
from hypothesis import note
from torch.testing import assert_close

import interp.circuit.testing.strategies as st_c
import rust_circuit as rc
import rust_circuit.module_library as mod_l
from interp.circuit.print_circuit import PrintCircuit
from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
from interp.circuit.testing.utils import mark_not_interesting_if, rand_matcher
from interp.tools.indexer import I
from rust_circuit import (
    Add,
    Array,
    Circuit,
    Concat,
    ConstructConcatShapeDifferentError,
    ConstructEinsumAxisSizeDifferentError,
    DiscreteVar,
    Einsum,
    ExpandConcatAxisError,
    Expander,
    ExpandFixedIndexError,
    ExpandGeneralFunctionTriedToBatchNonBatchableInputError,
    ExtractSymbolsBatchedInputError,
    ExtractSymbolsBoundInputInconsistentError,
    ExtractSymbolsHasBindingsFromOuterModuleError,
    GeneralFunction,
    Index,
    IterativeMatcher,
    Matcher,
    Module,
    ModuleArgSpec,
    ModulePusher,
    ModuleSpec,
    Parser,
    Rearrange,
    RearrangeSpec,
    RunDiscreteVarAllSpec,
    Sampler,
    Scalar,
    Shape,
    ShapeNotBroadcastableError,
    Symbol,
    TorchDeviceDtypeOp,
    all_children,
    any_children_with_symbolic_sizes,
    broadcast_shapes,
    cast_circuit,
    conform_all_modules,
    count_nodes,
    deep_map_preorder,
    default_nested_module_namer,
    elim_no_input_module,
    expand_node,
    extract_rewrite,
    extract_rewrite_raw,
    extract_symbols,
    extract_symbols_get,
    get_free_symbols,
    module_new_bind,
    module_remove_unused_inputs,
    new_traversal,
    replace_expand_bottom_up_dict,
    strip_names_and_tags,
    substitute_all_modules,
    symbolic_sizes,
)
from rust_circuit.interop_rust import rust_to_py
from rust_circuit.module_library import get_attention, get_bilinear_mlp, get_pointwise_mlp
from rust_circuit.py_utils import timed
from rust_circuit.rrfs import RRFS_DIR

from .test_rust_rewrite import get_c_st

P = Parser()
po = rc.PrintOptions(tensor_index_literal=True, bijection=False)


@hypothesis.given(get_c_st())
def test_subst_all_modules(circ):
    circ1 = rc.substitute_all_modules(circ)
    assert not any("internal_expand" in x.name for x in all_children(circ1))
    assert_close(circ.evaluate(), circ1.evaluate())


@hypothesis.given(get_c_st(max_growth_steps=10))
def test_conform_all_modules(circ):
    circ1 = rc.conform_all_modules(circ)
    assert not any("internal_expand" in x.name for x in all_children(circ1))
    assert_close(circ.evaluate(), circ1.evaluate())


@hypothesis.given(
    get_c_st(probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=30),
    st.data(),
)
@mark_not_interesting_if(rc.PushDownModulePushingPastModuleWhichOverridesSymError)
@mark_not_interesting_if(rc.ExpandModuleTriedToBatchUnbatchableInputError)
def test_push_down_modules(circ, d):
    skips = d.draw(st_c.st_subsets([x.name for x in all_children(circ) if x.is_module()]))
    cb = d.draw(st.builds(rc.ModulePusher.remove_and_elim_callback, st.booleans(), st.booleans(), st.booleans()))
    pu = rc.ModulePusher(flatten_modules=d.draw(st.booleans()), module_construct_callback=cb)
    note(circ.repr(po))
    circ1 = pu.push_down_modules(circ, traversal=rc.new_traversal(), skip_module=skips)
    note(circ1.repr(po))
    assert_close(circ.evaluate(), circ1.evaluate())


@pytest.mark.xfail
@hypothesis.given(
    get_c_st(probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=5),
    st.data(),
)
@mark_not_interesting_if(rc.ExtractSymbolsHasBindingsFromOuterModuleError)
def test_extract_symbols(circ, d):
    note(circ.repr(po))
    syms = d.draw(st_c.st_subsets([x for x in all_children(circ) if x.is_symbol()]))
    conform = d.draw(st.booleans())
    try:
        circ1 = rc.extract_symbols(
            circ, syms, use_elim_no_input_modules=d.draw(st.booleans()), conform_batch_if_needed=conform
        )
    except rc.ExtractSymbolsBatchedInputError:
        hypothesis.assume(conform)
        raise
    note(circ1.repr(po))
    assert_close(circ.evaluate(), circ1.evaluate())


@pytest.mark.xfail
@hypothesis.given(
    get_c_st(probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=10),
    st.data(),
)
def test_upd_nested_id(circ, d):
    note(circ.repr(po))
    try:
        circ1 = rc.update_bindings_nested(
            circ,
            lambda x: x if d.draw(st.booleans()) else None,
            rand_matcher(d),
            run_update_on_new_spec_circuits=d.draw(st.booleans()),
            flatten_modules=d.draw(st.booleans()),
        )
    except rc.ExceptionWithRustContext as e:
        raise e.exc
    assert_close(circ.evaluate(), circ1.evaluate())


def dedup_pairwise(l, f):
    out = []
    for x in l:
        if not any(f(x, y) for y in out):
            out.append(x)
    return out


@hypothesis.given(
    get_c_st(from_other=False, max_growth_steps=10, probs_per_depth=[CP.kw(all=0, branches=1, Cumulant=0)]).filter(
        lambda c: c.num_children > 0
    ),
    st.data(),
)
@mark_not_interesting_if(rc.RearrangeSpecNotConformableError)
@mark_not_interesting_if(rc.ExpandModuleTriedToBatchUnbatchableInputError)
def test_expander_prop(circ, d):
    children = list(rc.all_children(circ))
    children.sort()
    cs = d.draw(st.lists(st.sampled_from(children), unique=True, min_size=1, max_size=2))
    cs = dedup_pairwise(cs, lambda x, y: (x in rc.all_children(y) or (y in rc.all_children(x))))
    b = d.draw(st.integers(1, 5))
    # print("\ncirc:")
    # circ.print()
    # print("circs:")
    # for c in cs:
    #     c.print()
    new_cs = [d.draw(get_c_st(shape=(b,) + c.shape, from_other=False, max_growth_steps=4)) for c in cs]
    note(circ.repr(po))
    expanded = rc.Expander(*[(c, lambda _, new_c=new_c: new_c) for (c, new_c) in zip(cs, new_cs)])(circ)
    note(expanded.repr(po))
    d = {c: new_c for (c, new_c) in zip(cs, new_cs)}

    def upd(x, i):
        return x.update(set(cs), lambda a: rc.Index(d[a], (i,) + ((slice(None, None),) * a.ndim)))

    concatted = rc.Concat(*[upd(circ, i).unsqueeze(0) for i in range(b)], axis=0)
    note(concatted.repr(po))
    assert_close(expanded.evaluate(), concatted.evaluate())


def test_expand_node():
    circ: Circuit = Rearrange(
        Symbol.new_with_none_uuid((2, 3), "0"), RearrangeSpec([[0], [1]], [[1], [0]], [None, None])
    )
    expanded = expand_node(circ, [Array.randn(5, 6)])
    assert expanded.shape == (6, 5)
    print(expanded)

    circ: Circuit = Rearrange(
        Symbol.new_with_none_uuid((2, 3), "0"), RearrangeSpec([[0], [1]], [[1], [0]], [None, None])
    )
    expanded = expand_node(circ, [Array.randn(7, 5, 6)])
    assert expanded.shape == (7, 6, 5)
    print(expanded)

    circ: Circuit = Rearrange(
        Symbol.new_with_none_uuid((2, 3), "0"), RearrangeSpec([[0], [1]], [[1], [0]], [None, None])
    )
    expanded = expand_node(circ, [Array.randn(8, 7, 5, 6)])
    assert expanded.shape == (8, 7, 6, 5)
    print(expanded)

    circ = Index(Symbol.new_with_none_uuid((2, 3), "0"), I[:, 0])
    expanded = expand_node(circ, [Array.randn(5, 3)])
    assert expanded.shape == (5,)
    print(expanded)

    circ = Index(Symbol.new_with_none_uuid((2, 3), "0"), I[:, 0])
    expanded = expand_node(circ, [Array.randn(7, 5, 3)])
    assert expanded.shape == (7, 5)
    print(expanded)

    circ = Index(Symbol.new_with_none_uuid((2, 3), "0"), I[:, 0])
    expanded = expand_node(circ, [Array.randn(8, 7, 5, 3)])
    assert expanded.shape == (8, 7, 5)
    print(expanded)

    with pytest.raises(ExpandFixedIndexError):
        circ = Index(Symbol.new_with_none_uuid((2, 3), "0"), I[0:2, 0])
        expanded = expand_node(circ, [Array.randn(5, 3)])
        print(expanded)

    circ = Add(Symbol.new_with_none_uuid((2, 3, 1), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((7, 3, 1), "0"), Symbol.new_with_none_uuid((7, 1, 11), "1")]
    )
    assert expanded.shape == (7, 3, 11)
    print(expanded)

    circ = Add(Symbol.new_with_none_uuid((2, 3, 1), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((13, 7, 3, 1), "0"), Symbol.new_with_none_uuid((7, 1, 11), "1")]
    )
    assert expanded.shape == (13, 7, 3, 11)
    print(expanded)

    circ = Add(Symbol.new_with_none_uuid((2, 3, 1), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
    with pytest.raises(ShapeNotBroadcastableError):
        expand_node(
            circ, [Symbol.new_with_none_uuid((13, 7, 3, 1), "0"), Symbol.new_with_none_uuid((1, 7, 1, 11), "1")]
        )

    circ = Add(Symbol.new_with_none_uuid((2, 3, 1), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((4, 13, 7, 3, 1), "0"), Symbol.new_with_none_uuid((7, 1, 11), "1")]
    )
    assert expanded.shape == (4, 13, 7, 3, 11)
    print(expanded)

    circ = Add(Symbol.new_with_none_uuid((2, 3, 1), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((7, 3, 1), "0"), Symbol.new_with_none_uuid((4, 13, 7, 1, 11), "1")]
    )
    assert expanded.shape == (4, 13, 7, 3, 11)
    print(expanded)

    circ = Add(Symbol.new_with_none_uuid((3, 1), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((7, 3, 1), "0"), Symbol.new_with_none_uuid((8, 7, 9, 3, 1), "1")]
    )
    assert expanded.shape == (8, 7, 9, 3, 1)
    print(expanded)

    circ = Add(Symbol.new_with_none_uuid((), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
    with pytest.raises(ShapeNotBroadcastableError):
        expand_node(
            circ, [Symbol.new_with_none_uuid((7, 3, 1), "0"), Symbol.new_with_none_uuid((1, 3, 11, 9, 4, 5), "1")]
        )

    circ = Add(Symbol.new_with_none_uuid((), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
    with pytest.raises(ShapeNotBroadcastableError):
        expand_node(
            circ, [Symbol.new_with_none_uuid((1, 3, 1), "0"), Symbol.new_with_none_uuid((1, 3, 11, 9, 4, 5), "1")]
        )

    circ = Add(Symbol.new_with_none_uuid((7,), "0"), Symbol.new_with_none_uuid((2, 7), "1"))
    expanded = expand_node(circ, [Symbol.new_with_none_uuid((7, 7), "0"), Symbol.new_with_none_uuid((7, 7, 7), "1")])
    assert expanded.shape == (7, 7, 7)
    print(expanded)

    circ = Add(Symbol.new_with_none_uuid((8, 7), "0"), Symbol.new_with_none_uuid((1, 1, 1, 1, 1, 7), "1"))
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((7, 8, 9), "0"), Symbol.new_with_none_uuid((7, 1, 1, 1, 1, 1, 1), "1")]
    )
    assert expanded.shape == (7, 1, 1, 1, 1, 8, 9)
    print(expanded)

    circ = Add(Symbol.new_with_none_uuid((8, 7), "0"), Symbol.new_with_none_uuid((1, 1, 1, 1, 1, 7), "1"))
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((2, 4, 7, 8, 9), "0"), Symbol.new_with_none_uuid((7, 1, 1, 1, 1, 1, 1), "1")]
    )
    assert expanded.shape == (2, 4, 7, 1, 1, 1, 1, 8, 9)
    print(expanded)

    circ = Add(Symbol.new_with_none_uuid((8, 7), "0"), Symbol.new_with_none_uuid((1, 1, 1, 1, 1, 7), "1"))
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((7, 8, 9), "0"), Symbol.new_with_none_uuid((2, 4, 7, 1, 1, 1, 1, 1, 1), "1")]
    )

    circ = Einsum.from_einsum_string(
        "ab,bc->ac", Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")
    )
    expanded = expand_node(circ, [Symbol.new_with_none_uuid((5, 1, 2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")])
    assert expanded.shape == (5, 1, *circ.shape)
    print(expanded)

    circ = Einsum.from_einsum_string(
        "ab,bc->ac", Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")
    )
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((5, 1, 2, 3), "0"), Symbol.new_with_none_uuid((5, 1, 3, 5), "1")]
    )
    assert expanded.shape == (5, 1, *circ.shape)
    print(expanded)

    circ = Einsum.from_einsum_string(
        "ab,bc->ac", Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")
    )
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((5, 7, 2, 3), "0"), Symbol.new_with_none_uuid((7, 3, 5), "1")]
    )
    assert expanded.shape == (5, 7, *circ.shape)
    print(expanded)

    circ = Einsum.from_einsum_string(
        "ab,bc->ac", Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")
    )
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((5, 7, 2, 12), "0"), Symbol.new_with_none_uuid((7, 12, 11), "1")]
    )
    assert expanded.shape == (5, 7, 2, 11)
    print(expanded)

    with pytest.raises(ConstructEinsumAxisSizeDifferentError):
        circ = Einsum.from_einsum_string(
            "ab,bc->ac", Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")
        )
        expanded = expand_node(
            circ, [Symbol.new_with_none_uuid((5, 2, 2, 2), "0"), Symbol.new_with_none_uuid((5, 2, 3, 5), "1")]
        )
        print(expanded)

    with pytest.raises(ShapeNotBroadcastableError):
        circ = Einsum.from_einsum_string(
            "ab,bc->ac", Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")
        )
        expanded = expand_node(
            circ, [Symbol.new_with_none_uuid((5, 3, 2, 3), "0"), Symbol.new_with_none_uuid((5, 2, 3, 5), "1")]
        )

    with pytest.raises(ExpandConcatAxisError):
        circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), axis=0)
        expanded = expand_node(circ, [Symbol.new_with_none_uuid((3, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1")])

    with pytest.raises(ShapeNotBroadcastableError):
        circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 3), "1"), axis=0)
        expanded = expand_node(
            circ, [Symbol.new_with_none_uuid((2, 2, 3), "0"), Symbol.new_with_none_uuid((3, 3, 3), "1")]
        )

    with pytest.raises(ConstructConcatShapeDifferentError):
        circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 3), "1"), axis=0)
        expanded = expand_node(
            circ, [Symbol.new_with_none_uuid((3, 2, 2), "0"), Symbol.new_with_none_uuid((3, 3, 3), "1")]
        )

    circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), axis=0)
    expanded = expand_node(circ, [Symbol.new_with_none_uuid((5, 2, 3), "0"), Symbol.new_with_none_uuid((5, 4, 3), "1")])
    assert expanded.shape == (5, *circ.shape)
    print(expanded)

    circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), axis=0)
    expanded = expand_node(circ, [Symbol.new_with_none_uuid((5, 2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1")])
    assert expanded.shape == (5, *circ.shape)
    print(expanded)

    circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), axis=0)
    expanded = expand_node(circ, [Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((7, 3, 4, 3), "1")])
    assert expanded.shape == (7, 3, *circ.shape)
    print(expanded)

    circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), axis=0)
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((3, 2, 3), "0"), Symbol.new_with_none_uuid((7, 3, 4, 3), "1")]
    )
    assert expanded.shape == (7, 3, *circ.shape)
    print(expanded)

    with pytest.raises(ExpandGeneralFunctionTriedToBatchNonBatchableInputError):
        circ = GeneralFunction.gen_index(
            Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), index_dim=-1
        )
        expanded = expand_node(
            circ, [Symbol.new_with_none_uuid((3, 2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1")]
        )

    circ = GeneralFunction.gen_index(
        Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), index_dim=-1
    )
    expanded = expand_node(circ, [Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((7, 4, 3), "1")])
    assert expanded.shape == (7, 4, 3, 2)

    circ = GeneralFunction.gen_index(
        Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), index_dim=-1
    )
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((7, 11, 4, 3), "1")]
    )
    assert expanded.shape == (7, 11, 4, 3, 2)

    circ = GeneralFunction.gen_index(
        Symbol.new_with_none_uuid((4, 3, 2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), index_dim=-1, batch_x=True
    )
    expanded = expand_node(
        circ, [Symbol.new_with_none_uuid((4, 3, 2, 3), "0"), Symbol.new_with_none_uuid((7, 11, 4, 3), "1")]
    )
    expanded.get_unique("0 rep_for_batch")  # TODO: maybe rename me!
    assert expanded.shape == (7, 11, 4, 3, 2)

    s = """
    'rep_fst' [?, ?] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
    'rep_snd' [?, ?] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
    'a' Module
      'add' Add
        'arg0' [] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
        'arg1' [] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
      'fst' [?, ?] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
      'snd' [?, ?] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
    """
    rep_fst, rep_snd, c = Parser().parse_circuits(s)
    c = c.cast_module()
    out = expand_node(c, [c.spec.circuit, rep_fst, rep_snd])
    assert out.shape[0] in [rep_fst.shape[0], rep_snd.shape[0]]
    assert out.shape[1] in [rep_fst.shape[1], rep_snd.shape[1]]

    s = """
    'rep_fst' [?] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
    'rep_snd' [8, ?] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
    'a' Module
      'add' Add
        'arg0' [] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
        'arg1' [] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
      'fst' [2] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
      'snd' [3, 2] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
    """
    rep_fst, rep_snd, c = Parser().parse_circuits(s)
    c = c.cast_module()
    out = expand_node(c, [c.spec.circuit, rep_fst, rep_snd])
    assert out.shape[0] in [rep_snd.shape[0]]
    assert out.shape[1] in [rep_fst.shape[0], rep_snd.shape[1]]

    s = """
    'rep_fst' [7, 9, ?] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
    'rep_snd' [8, ?] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
    'a' Module
      'add' Add
        'arg0' [?] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
        'arg1' [?] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
      'fst' [?] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
      'snd' [?, ?] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
    'b' Module
      'a'
      'rep_fst' ! 'fst'
      'rep_snd' ! 'snd'
    """
    rep_fst, rep_snd, c, b_mod = Parser().parse_circuits(s)
    assert b_mod.shape[:3] == (7, 9, 8)
    assert b_mod.shape[3] in [rep_fst.shape[-1], rep_snd.shape[-1]]
    sub = b_mod.cast_module().substitute()
    assert sub.shape[:3] == (7, 9, 8)
    assert sub.shape[3] in [rep_fst.shape[-1], rep_snd.shape[-1]]

    for sym_on_top in [False, True]:
        s = f"""
        'rep_fst' [12, {"?" if sym_on_top else 8}, 9, 7] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
        'rep_snd' [13, 12, 9, 10, {8 if sym_on_top else "?"}, 7] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
        'a' Module
          'add' Add
            'arg0' [?, ?] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
            'arg1' [?] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
          'fst' [?, ?, ?] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
          'snd' [?, ?, ?, ?] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
        'b' Module
          'a'
          'rep_fst' ! 'fst'
          'rep_snd' ! 'snd'
        """
        rep_fst, rep_snd, c, out = Parser().parse_circuits(s)
        out.print()
        out.cast_module().substitute().print()
        assert out.shape == (13, 12, 9, 10, 8, 9, 7)
        assert out.cast_module().substitute().shape == (13, 12, 9, 10, 8, 9, 7)

    s = """
    'rep_fst' [    12,      102, 7] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
    'rep_snd' [13, 12,  10, 102, 7] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
    'a' Module
      'add' Add
        'arg0' [?] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
        'arg1' [?] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
      'fst' [   ?, ?] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
      'snd' [?, ?, ?] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
    'b' Module
      'a'
      'rep_fst' ! 'fst'
      'rep_snd' ! 'snd'
    """
    rep_fst, rep_snd, c, out = Parser().parse_circuits(s)
    out.print()
    out.cast_module().substitute().print()
    assert out.shape == (13, 12, 10, 102, 7)
    assert out.cast_module().substitute().shape == (13, 12, 10, 102, 7)


def test_module_shape_missing():
    s0, s1, *_ = symbolic_sizes()
    spec = ModuleSpec(
        Scalar(0.0),
        [ModuleArgSpec(Symbol.new_with_none_uuid((s0,), "a")), ModuleArgSpec(Symbol.new_with_none_uuid((s1,), "b"))],
        check_all_inputs_used=False,
    )

    c = Module(spec, a=Symbol.new_with_none_uuid((4, 5, 6, 7)), b=Symbol.new_with_none_uuid((6, 4)))

    assert c.shape == (4, 5, 6)
    assert c.substitute().shape == (4, 5, 6)
    c.substitute().print()

    c = Module(spec, a=Symbol.new_with_none_uuid((4, 5, 6, 7)), b=Symbol.new_with_none_uuid((4,)))
    assert c.shape == (4, 5, 6)
    assert c.substitute().shape == (4, 5, 6)
    c.substitute().print()

    c = Module(spec, a=Symbol.new_with_none_uuid((7,)), b=Symbol.new_with_none_uuid((4, 5, 6, 4)))
    assert c.shape == (4, 5, 6)
    assert c.substitute().shape == (4, 5, 6)
    c.substitute().print()

    spec = ModuleSpec(
        Scalar(0.0, shape=(7,)),
        [ModuleArgSpec(Symbol.new_with_none_uuid((s0,), "a")), ModuleArgSpec(Symbol.new_with_none_uuid((s1,), "b"))],
        check_all_inputs_used=False,
    )

    c = Module(spec, a=Symbol.new_with_none_uuid((7,)), b=Symbol.new_with_none_uuid((4, 5, 6, 4)))
    assert c.shape == (4, 5, 6, 7)
    assert c.substitute().shape == (4, 5, 6, 7)
    c.substitute().print()


def test_module_nodes():
    spec = ModuleSpec(
        Einsum.from_einsum_string(
            "ab,bc->ac", Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")
        ),
        [
            ModuleArgSpec(Symbol.new_with_none_uuid((2, 3), "0"), True, True),
            ModuleArgSpec(Symbol.new_with_none_uuid((3, 5), "1"), True, True),
        ],
    )
    modulenode = Module.new_flat(spec, Array.randn(2, 3), Array.randn(3, 5))
    print((modulenode))


def test_module_nodes_2():
    spec = get_pointwise_mlp().spec
    spec.circuit.print()
    modulenode = Module.new_flat(
        spec, Array.randn(256), Array.randn(256 * 4, 256), Array.randn(256 * 4), Array.randn(256, 256 * 4)
    )
    modulenode.print()

    spec = get_bilinear_mlp(output_bias=False).spec
    spec.circuit.print()
    modulenode = Module(
        spec,
        "modulename",
        **{
            "m.input": Array.randn(256, name="inputy"),
            "m.w.proj_in": Array.randn(256 * 4, 256, name="project_in_y"),
            "m.w.in_bias": Array.randn(256 * 4, name="bias_in_y"),
            "m.w.proj_out": Array.randn(256, 256 * 2, name="project_out_y"),
        },
    )
    modulenode.print()
    modulenode.substitute().print()

    spec = get_attention().spec
    spec.circuit.print()
    modulenode = Module(
        spec,
        **{
            "a.w.q": Array.randn(4, 16, 64),
            "a.w.k": Array.randn(4, 16, 64),
            "a.w.v": Array.randn(4, 16, 64),
            "a.w.o": Array.randn(4, 64, 16),
            "a.mask": Array.randn(5, 5),
            "a.input": Array.randn(5, 64),
        },
    )
    modulenode.print()
    modulenode.substitute().print()
    substitute_all_modules(modulenode).print()


def test_bind():
    proj_in, in_bias, proj_out, inp = Array.randn(64, 16), Array.randn(64), Array.randn(16, 64), Array.randn(16)
    gelu_mlp_free_body = get_pointwise_mlp().body
    m = module_new_bind(gelu_mlp_free_body, ("m.w.proj_in", proj_in))
    m = module_new_bind(m, ("m.w.proj_out", proj_out), ("m.input", inp))
    m = module_new_bind(m, ("m.w.in_bias", in_bias))
    torch.testing.assert_close(
        m.evaluate(),
        Module(
            ModuleSpec.new_free_symbols(gelu_mlp_free_body),
            name="g",
            **dict([("m.w.proj_in", proj_in), ("m.w.in_bias", in_bias), ("m.w.proj_out", proj_out), ("m.input", inp)]),
        ).evaluate(),
    )


def test_replace_deep():
    replace_expand_bottom_up_dict(
        Einsum.from_einsum_string(
            "ab,bc->ac",
            Add(
                Symbol.new_with_none_uuid((2, 3), "0"),
                Symbol.new_with_none_uuid((2, 3), "0"),
            ),
            Symbol.new_with_none_uuid((3, 5), "1"),
        ),
        {Symbol.new_with_none_uuid((2, 3), "0"): Symbol.new_with_none_uuid((100, 3), "0")},
    ).print()


def test_extract():
    ispecs = [
        (Array.randn(2, 3), ModuleArgSpec(Symbol.new_with_none_uuid((2, 3), "0"), True, True)),
        (Array.randn(3, 5), ModuleArgSpec(Symbol.new_with_none_uuid((3, 5), "1"), True, True)),
    ]
    circ = Einsum.from_einsum_string(
        "ab,bc->ac",
        Add(
            ispecs[0][0],
            ispecs[0][0],
        ),
        ispecs[1][0],
    )
    circ.print()
    module_spec = ModuleSpec.new_extract(circ, cast(Any, ispecs))
    module_spec.circuit.print()


@pytest.mark.skip("rrfs stuff lazy ci")
def test_extract_gpt():
    with timed("load"):
        orig = Parser(on_repeat_check_info_same=False)(
            open(RRFS_DIR + "/tao/circuits/gpt2_small_300_blank.circ").read()
        )
    py_version = rust_to_py(orig)
    PrintCircuit(print_html=True)(py_version)
    # with timed("print"):
    #     orig.print()
    cur = orig
    for i in range(12):
        matcher = Matcher.any(Matcher(Array), Matcher(f"a{i}.inp"))
        children = list(matcher.get(cur))
        # print(children)
        children.sort(key=lambda x: x.name)
        cur = Matcher(f"a{i}").update(
            cur,
            lambda x: extract_rewrite_raw(
                x, [(c, ModuleArgSpec.just_name_shape(c)) for c in children], f"a{i}.", f"a{i}"
            ),
            fancy_validate=True,
        )
        Matcher(f"a{i}").update(
            cur,
            lambda x: extract_rewrite(
                x,
                Matcher.any(Matcher(Array), Matcher(f"a{i}.inp")),
                f"a{i}.",
                f"a{i}",
                circuit_to_arg_spec=ModuleArgSpec.just_name_shape,
            ),
            fancy_validate=True,
        )
        # print("Traversal")
        # cur_2.print()
        # print("Raw")
        # cur.print()
        # print(cur.hash, cur_2.hash)
        # torch.testing.assert_close(cur.evaluate(), cur_2.evaluate())
        # # assert cur == cur_2

        matcher = Matcher.any(Matcher.all(Matcher(r"^m\d."), Matcher(Array)), Matcher(f"m{i}.inp"))
        children = list(matcher.get(cur))
        children.sort(key=lambda x: x.name)
        cur = Matcher(f"m{i}").update(
            cur,
            lambda x: extract_rewrite_raw(
                x,
                [(c, ModuleArgSpec.just_name_shape(c)) for c in children],
                f"m{i}.",
                f"m{i}",
            ),
            fancy_validate=True,
        )
    cur.print()
    a0s = Matcher("a0").get_unique(cur).cast_module().spec.circuit
    a0s.print()
    a1s = Matcher("a1").get_unique(cur).cast_module().spec.circuit
    a1s.print()
    print(a0s == a1s)


def raw_test_full_transformer(
    params: mod_l.TransformerBlockParams,
):
    t, _, _ = mod_l.TransformerParams(params, num_layers=7).garbage_call(
        head_size=3,
        num_heads=5,
        seq_len=11,
        batch_shape=(2, 3),
    )
    with timed("conform"):
        conformed = conform_all_modules(t)
    assert not any_children_with_symbolic_sizes(conformed)
    with timed("preorder inline"):
        deep_map_preorder(t, lambda x: x.substitute() if isinstance(x, Module) else x)
    t = strip_names_and_tags(t)
    with timed("inline"):
        t = substitute_all_modules(t)
    print(count_nodes(t))


def test_full_transformer():
    Params = mod_l.TransformerBlockParams
    raw_test_full_transformer(Params("ln", False, False, True, "gelu", False))
    raw_test_full_transformer(Params("ln", False, False, True, "relu", False))
    raw_test_full_transformer(Params("ln", False, False, True, "bilinear", False))
    raw_test_full_transformer(Params("ln", False, False, True, "relu", True))
    raw_test_full_transformer(Params("ln", False, False, True, "bilinear", True))
    raw_test_full_transformer(Params(None, False, False, True, "gelu", False))
    raw_test_full_transformer(Params("ln", True, True, True, "gelu", False))
    raw_test_full_transformer(Params("ln", False, True, True, "gelu", False))
    raw_test_full_transformer(Params("ln", False, False, False))
    raw_test_full_transformer(Params(None, False, False, False))
    raw_test_full_transformer(Params("bn", False, False, True, "gelu", False))
    raw_test_full_transformer(Params("bn", False, False, False))


def test_module_conform():
    spec = get_attention().spec
    module = Module(
        spec,
        **{
            "a.w.q": Array.randn(4, 16, 64),
            "a.w.k": Array.randn(4, 16, 64),
            "a.w.v": Array.randn(4, 16, 64),
            "a.w.o": Array.randn(4, 64, 16),
            "a.mask": Array.randn(5, 5),
            "a.input": Array.randn(5, 64),
        },
    )
    module.print()
    new_module = module.conform_to_input_shapes()
    new_module.print()

    torch.testing.assert_close(module.evaluate(), new_module.evaluate())


@pytest.mark.parametrize("separate_mod", [False, True])
def test_complex_batching(separate_mod: bool):
    r = partial(get_complex_module_test, separate_mod=separate_mod)

    r()
    r(batch_inner_0=(3,))
    r(batch_inner_0=(3, 7))
    r(batch_inner_1=(3,))
    r(batch_inner_1=(3, 7))
    r(batch_inner_0=(3,), batch_inner_1=(3,))
    r(batch_inner_0=(7,), batch_inner_1=(7,))
    r(batch_inner_0=(3,), batch_inner_1=(7, 3))
    r(batch_inner_0=(7, 3), batch_inner_1=(7, 3))
    r(batch_inner_0=(7, 3), batch_inner_1=(3,))
    r(batch_inner_0=(1, 7, 3), batch_inner_1=(3,))

    r(batch_outer_0=(3,))
    r(batch_outer_0=(3, 7))
    r(batch_outer_1=(3,))
    r(batch_outer_1=(3, 7))
    r(batch_outer_0=(3,), batch_outer_1=(3,))
    r(batch_outer_0=(7,), batch_outer_1=(7,))
    r(batch_outer_0=(3,), batch_outer_1=(7, 3))
    r(batch_outer_0=(7, 3), batch_outer_1=(7, 3))
    r(batch_outer_0=(7, 3), batch_outer_1=(3,))
    r(batch_outer_0=(1, 7, 3), batch_outer_1=(3,))

    r(batch_inner_0=(3,), batch_outer_0=(3,))
    r(batch_inner_0=(3,), batch_outer_0=(4,))
    r(batch_inner_0=(3, 7), batch_outer_0=(4,), batch_outer_1=(6, 4))
    _, _, outer = r(batch_inner_0=(3, 7), batch_inner_1=(7,), batch_outer_0=(4,), batch_outer_1=(6, 4))
    outer.cast_module().substitute().print()
    r(batch_inner_1=(7,), batch_outer_0=(4,), batch_outer_1=(6, 4))
    r(batch_inner_1=(3,), batch_outer_0=(4,))
    r(batch_inner_1=(3,), batch_outer_1=(4,))
    r(batch_inner_0=(3,), batch_outer_1=(4,))
    r(batch_inner_0=(3,), batch_outer_1=(6, 4))


def get_complex_module_test(
    batch_inner_0: Shape = (),
    batch_inner_1: Shape = (),
    batch_outer_0: Shape = (),
    batch_outer_1: Shape = (),
    separate_mod: bool = False,
):
    alpha = [chr(i) for i in range(ord("f"), ord("z"))]
    b_inner_alpha = "".join(alpha[: len(batch_inner_0)])
    s = f"""
    'inner' Einsum az,bz,cl->bac
      'inner.sym_arg0' [?, ?] Symbol a409cdc3-d962-4df1-8d1b-c5c8d8b50639
      'inner.concat' Concat 0
        'inner.sym_arg1' [3, ?] Symbol 02927290-fc4f-45b7-9822-267918e897d6
        'outer.sym_arg0' [7, ?] Symbol 6dcfaf1e-5e35-418a-ad88-3d79ae32096c
      'outer.sym_arg1' [?, ?] Symbol ecf273f5-82c5-4fbf-8097-d9b1a250e035

    'inner.call' Module
      'inner'
      'inner.apply_inner' Einsum cd,{b_inner_alpha}a->{b_inner_alpha}ca ! 'inner.sym_arg0'
        'outer.sym_arg0'
        'inner.rand0' [{",".join(map(str, batch_inner_0 + (5,)))}] Array rand
      'inner.rand1' [{",".join(map(str, batch_inner_1 + (3, 5)))}] Array rand ! 'inner.sym_arg1'

    'add_inner' Add
      'inner.call'
      'inner.call'

    'outer' Module
      '{"add_inner" if separate_mod else "inner.call"}'
      'outer.rand0' [{",".join(map(str, batch_outer_0 + (7, 5)))}] Array rand ! 'outer.sym_arg0'
      'outer.rand1' [{",".join(map(str, batch_outer_1 + (9, 12)))}] Array rand ! 'outer.sym_arg1'
    """

    inner, inner_call_v, _, outer_v = [
        cast_circuit(c, TorchDeviceDtypeOp(dtype="float64")) for c in Parser().parse_circuits(s)
    ]
    inner_call = inner_call_v.cast_module()
    outer = outer_v.cast_module()
    inner_batch = broadcast_shapes(batch_inner_0, batch_inner_1)
    outer_batch = broadcast_shapes(batch_outer_0, batch_outer_1)
    assert inner.shape == (10, rc.Size.NONE, rc.Size.NONE)
    assert inner_call.shape == (inner_batch + (10, 7, rc.Size.NONE))
    assert outer.shape == (outer_batch + inner_batch + (10, 7, 9))

    base_value = outer.evaluate()

    def test_close(outer: Circuit):
        def sub(x: Circuit):
            out = x.cast_module().substitute()
            assert out.name == x.name
            return out

        torch.testing.assert_close(
            substitute_all_modules(outer).evaluate(), base_value
        )  # does same as base under current impl
        torch.testing.assert_close(
            outer.update("inner.call", sub).evaluate(), base_value
        )  # does same as base under current impl
        torch.testing.assert_close(sub(outer).evaluate(), base_value)

    test_close(outer)

    def get_rand_batcher(n: str, batch_size: int):
        probs = DiscreteVar.uniform_probs_and_group(batch_size, name="probs")

        def get_discrete_var_circ(s: str):
            circ = outer.get_unique(s)
            return DiscreteVar(circ, probs, name=circ.name + " var")

        return (
            Expander(*[(s, partial(lambda s, _: get_discrete_var_circ(s), s)) for s in [f"{n}.rand0", f"{n}.rand1"]]),
            probs,
        )

    if batch_inner_0 != () or batch_inner_1 != () or batch_outer_0 != () or batch_outer_1 != ():
        no_batch_inner, no_batch_inner_call, no_batch_outer = get_complex_module_test(separate_mod=separate_mod)

        def get_manual_batcher(n: str):
            return Expander(*[(s, partial(lambda s, _: outer.get_unique(s), s)) for s in [f"{n}.rand0", f"{n}.rand1"]])

        def batch(x: Circuit):
            return get_manual_batcher("outer")(get_manual_batcher("inner")(x))

        assert batch(no_batch_inner).shape == inner.shape
        assert batch(no_batch_inner_call).shape == inner_call.shape
        assert batch(no_batch_outer).shape == outer.shape

        test_close(batch(no_batch_outer))

        if batch_inner_0 == batch_inner_1 and len(batch_inner_0) == 1 and batch_outer_0 == () and batch_outer_1 == ():
            outer_batched = get_manual_batcher("outer")(no_batch_outer)
            batcher, probs = get_rand_batcher("inner", batch_inner_0[0])
            inner_rand = batcher(outer_batched)
            sampled = Sampler(RunDiscreteVarAllSpec([probs])).sample(inner_rand)
            torch.testing.assert_close(sampled.evaluate(), base_value)

        if batch_outer_0 == batch_outer_1 and len(batch_outer_0) == 1 and batch_inner_0 == () and batch_inner_1 == ():
            inner_batched = get_manual_batcher("inner")(no_batch_outer)
            batcher, probs = get_rand_batcher("outer", batch_outer_0[0])
            outer_rand = batcher(inner_batched)
            sampled = Sampler(RunDiscreteVarAllSpec([probs])).sample(outer_rand)
            torch.testing.assert_close(sampled.evaluate(), base_value)

    return inner, inner_call, outer


def test_elim_no_input_module():
    s = """
    'inside' [] Scalar 4
    'a' Module
      'inside'
    """
    inside, circ = Parser().parse_circuits(s)
    assert elim_no_input_module(circ.cast_module()) == inside

    s = """
    'inside' [] Scalar 4
    'a' Module
      'inside'
      'a.bound' [] Scalar 3 ! 'a.inp' [] Symbol 90e7ce1a-0d52-4845-b42e-82411d6cbf56

    """
    inside, circ = Parser(module_check_all_inputs_used=False).parse_circuits(s)
    assert elim_no_input_module(circ.cast_module()) is None


def test_remove_unused_inputs():
    s = """
    'inside' [] Scalar 4
    0 'a rem_unused' Module
      'inside'
    1 'a' Module
      'inside'
      'a.bound' [] Scalar 3 ! 'a.inp' [] Symbol 90e7ce1a-0d52-4845-b42e-82411d6cbf56

    """
    inside, removed_circ, circ = Parser(module_check_all_inputs_used=False).parse_circuits(s)
    assert module_remove_unused_inputs(circ.cast_module(), use_elim_no_input_module=True) == inside
    assert module_remove_unused_inputs(circ.cast_module(), use_elim_no_input_module=False) == removed_circ

    s = """
    'inside' [] Scalar 4
    1 'a' Module
      'inside'
      'a.bound' [2, 3] Scalar 3 ! 'a.inp' [] Symbol 90e7ce1a-0d52-4845-b42e-82411d6cbf56

    """
    inside, circ = Parser(module_check_all_inputs_used=False).parse_circuits(s)
    out_removed = module_remove_unused_inputs(circ.cast_module(), use_elim_no_input_module=True)
    assert len(out_removed.get(Matcher.types(Module, Symbol))) == 0
    assert out_removed.shape == (2, 3)
    assert (out_removed.evaluate() == circ.evaluate()).all()
    out_strip_no_drop = module_remove_unused_inputs(circ.cast_module(), use_elim_no_input_module=False)
    assert len(out_strip_no_drop.get(Matcher.types(Symbol))) == 0
    assert len(out_strip_no_drop.get(Matcher.types(Module))) == 1
    assert out_removed.shape == (2, 3)
    assert (out_removed.evaluate() == circ.evaluate()).all()

    s = """
    'a' Module
      'stuff' Add
        'a.inp' [] Symbol 90e7ce1a-0d52-4845-b42e-82411d6cbf56
        'other' [3] Scalar 5
      'a.bound' [2, 3] Scalar 3 ! 'a.inp'
    """
    circ = Parser().parse_circuit(s)
    assert module_remove_unused_inputs(circ.cast_module(), use_elim_no_input_module=True) == circ
    assert module_remove_unused_inputs(circ.cast_module(), use_elim_no_input_module=False) == circ

    fancy_test_remove_unused_inputs()
    fancy_test_remove_unused_inputs(unbound0_batch=(2,))
    fancy_test_remove_unused_inputs(unbound1_batch=(2,))
    fancy_test_remove_unused_inputs(unbound0_batch=(2,), unbound1_batch=(2,))
    fancy_test_remove_unused_inputs(unbound0_batch=(3, 2), unbound1_batch=(2,))
    fancy_test_remove_unused_inputs(unbound0_batch=(2,), unbound1_batch=(3, 2))
    fancy_test_remove_unused_inputs(unbound0_batch=(4, 3, 2), unbound1_batch=(2,))
    fancy_test_remove_unused_inputs(unbound0_batch=(2,), unbound1_batch=(4, 3, 2))

    fancy_test_remove_unused_inputs(bound0_batch=(2,))
    fancy_test_remove_unused_inputs(bound1_batch=(2,))
    fancy_test_remove_unused_inputs(bound2_batch=(2,))
    fancy_test_remove_unused_inputs(bound0_batch=(2,), bound1_batch=(2,))
    fancy_test_remove_unused_inputs(bound0_batch=(3, 2), bound1_batch=(2,))
    fancy_test_remove_unused_inputs(bound0_batch=(2,), bound1_batch=(3, 2))
    fancy_test_remove_unused_inputs(bound0_batch=(4, 3, 2), bound1_batch=(2,))
    fancy_test_remove_unused_inputs(bound0_batch=(2,), bound1_batch=(4, 3, 2))

    fancy_test_remove_unused_inputs(
        bound0_batch=(2,), bound1_batch=(4, 3, 2), unbound0_batch=(2,), unbound1_batch=(4, 3, 2)
    )
    fancy_test_remove_unused_inputs(
        bound0_batch=(2,), bound1_batch=(4, 3, 2), bound2_batch=(3, 2), unbound0_batch=(2,), unbound1_batch=(4, 3, 2)
    )
    fancy_test_remove_unused_inputs(
        bound0_batch=(2,), bound1_batch=(3, 2), bound2_batch=(2,), unbound0_batch=(2,), unbound1_batch=(4, 3, 2)
    )
    fancy_test_remove_unused_inputs(
        bound0_batch=(2,), bound1_batch=(3, 2), bound2_batch=(2,), unbound0_batch=(4, 3, 2), unbound1_batch=(2,)
    )


def fancy_test_remove_unused_inputs(
    unbound0_batch: Shape = (),
    unbound1_batch: Shape = (),
    bound0_batch: Shape = (),
    bound1_batch: Shape = (),
    bound2_batch: Shape = (),
):
    s = f"""
    'a' Module
      2 Einsum i,i,,ij->i
        'a.inp_bound0' [4] Symbol 7f6cd401-b2f0-4b2d-9967-4b0c2692bdfc
        'stuff' [4] Array rand
        'a.inp_bound1' [] Symbol e4c8dcf0-e3fb-44d7-857a-60c839b216cb
        'a.inp_bound2' [4, 5] Symbol 9be09e34-f084-435e-9297-51ffce4843ca

      'a.unbound0' [{",".join(map(str, unbound0_batch))}] Array rand ! 'a.inp_unbound0' [] Symbol 90e7ce1a-0d52-4845-b42e-82411d6cbf56
      'a.bound0' [{",".join(map(str, bound0_batch + (4,)))}] Array rand ! 'a.inp_bound0' [4] Symbol 7f6cd401-b2f0-4b2d-9967-4b0c2692bdfc
      'a.unbound1' [{",".join(map(str, unbound1_batch + (6, 7)))}] Array rand ! 'a.inp_unbound1' [6, 7] Symbol a87b33e7-85bc-412c-b8df-708821df0598
      'a.bound1' [{",".join(map(str, bound1_batch ))}] Array rand ! 'a.inp_bound1' [] Symbol e4c8dcf0-e3fb-44d7-857a-60c839b216cb
      'a.bound2' [{",".join(map(str, bound2_batch + (4, 5)))}] Array rand ! 'a.inp_bound2' [4, 5] Symbol 9be09e34-f084-435e-9297-51ffce4843ca
    """
    circ = Parser(module_check_all_inputs_used=False).parse_circuit(s)
    out_strip = module_remove_unused_inputs(circ.cast_module(), use_elim_no_input_module=True)
    out_strip.print()
    assert out_strip == module_remove_unused_inputs(circ.cast_module(), use_elim_no_input_module=False)
    assert len(out_strip.get(Matcher.types(Symbol))) == 3
    torch.testing.assert_close(out_strip.evaluate(), circ.evaluate())


def run_test_push_down_mod(
    batch_inner_0: Shape = (),
    batch_inner_1: Shape = (),
    batch_outer_0: Shape = (),
    batch_outer_1: Shape = (),
    separate_mod: bool = False,
):
    _, inner_call, outer = get_complex_module_test(
        batch_inner_0=batch_inner_0,
        batch_inner_1=batch_inner_1,
        batch_outer_0=batch_outer_0,
        batch_outer_1=batch_outer_1,
        separate_mod=separate_mod,
    )
    outer.print()

    # up to some minor difference in rearrange matching, push down is equiv to all sub
    push = ModulePusher().push_down_modules(outer, traversal=new_traversal())
    push.print()
    assert push.get(Symbol) == set()
    torch.testing.assert_close(push.evaluate(), outer.evaluate())

    for flatten_modules, remove_unused_inputs, elim_no_input_modules in itertools.product(*([[False, True]] * 3)):
        # none of these args should effect this case
        equiv_push = ModulePusher(
            flatten_modules=flatten_modules,
            module_construct_callback=ModulePusher.remove_and_elim_callback(
                remove_unused_inputs=remove_unused_inputs,
                elim_no_input_modules=elim_no_input_modules,
            ),
        ).push_down_modules(outer, traversal=new_traversal())
        assert equiv_push == push

    push = ModulePusher(flatten_modules=True).push_down_modules(
        outer, traversal=new_traversal(term_early_at="inner.concat")
    )
    push.print()
    assert len(push.get(Symbol)) == 2
    assert len(push.get(Module)) == 1
    assert sum(len(m.cast_module().spec.arg_specs) for m in push.get(Module)) == 2
    torch.testing.assert_close(push.evaluate(), outer.evaluate())

    push = ModulePusher(flatten_modules=False).push_down_modules(
        outer, traversal=new_traversal(term_early_at="inner.concat")
    )
    push.print()
    assert len(push.get(Symbol)) == 2
    assert len(push.get(Module)) == 2
    assert sum(len(m.cast_module().spec.arg_specs) for m in push.get(Module)) == 2
    torch.testing.assert_close(push.evaluate(), outer.evaluate())

    push = ModulePusher(
        flatten_modules=False, module_construct_callback=ModulePusher.noop_callback()
    ).push_down_modules(outer, traversal=new_traversal(term_early_at="inner.concat"))
    push.print()
    assert len(push.get(Symbol)) == 2
    assert len(push.get(Module)) == 2
    assert sum(len(m.cast_module().spec.arg_specs) for m in push.get(Module)) == 4
    torch.testing.assert_close(push.evaluate(), outer.evaluate())

    push = ModulePusher(flatten_modules=True).push_down_modules(
        outer, traversal=new_traversal(term_early_at="inner.concat")
    )
    push.print()
    assert len(push.get(Symbol)) == 2
    assert len(push.get(Module)) == 1
    assert sum(len(m.cast_module().spec.arg_specs) for m in push.get(Module)) == 2
    torch.testing.assert_close(push.evaluate(), outer.evaluate())

    assert (
        ModulePusher().push_down_modules(outer, traversal=new_traversal(), skip_module=IterativeMatcher(True)) == outer
    )
    assert (
        ModulePusher().push_down_modules(outer, traversal=new_traversal(), skip_module={"outer", "inner.call"}) == outer
    )

    push_skip_inner = ModulePusher().push_down_modules(outer, traversal=new_traversal(), skip_module={"inner.call"})
    push_skip_inner.print()
    torch.testing.assert_close(push_skip_inner.evaluate(), outer.evaluate())

    push_skip_inner = ModulePusher().push_down_modules(
        outer, traversal=new_traversal(term_early_at="inner.concat"), skip_module={"inner.call"}
    )
    push_skip_inner.print()
    torch.testing.assert_close(push_skip_inner.evaluate(), outer.evaluate())

    push = ModulePusher().push_down_modules(outer, traversal=new_traversal(), skip_module={"outer"})
    push.print()
    assert isinstance(push, Module)
    assert push.spec.arg_specs == outer.spec.arg_specs
    assert len(push.get(Symbol)) == 2
    assert len(push.get(Module)) == 1
    torch.testing.assert_close(push.evaluate(), outer.evaluate())

    if batch_inner_1 == ():
        out = extract_symbols(
            outer, {sp.symbol for sp in inner_call.spec.arg_specs if sp.symbol.name == "inner.sym_arg1"}
        )
        out.print()
        torch.testing.assert_close(out.evaluate(), outer.evaluate())

        out_new = extract_symbols_get(outer, "inner.sym_arg1")
        assert out == out_new

    if len(batch_inner_1) > 0:
        with pytest.raises(ExtractSymbolsBatchedInputError):
            extract_symbols_get(outer, "inner.sym_arg1")

    if batch_inner_0 == ():
        with pytest.raises(ExtractSymbolsHasBindingsFromOuterModuleError):
            extract_symbols_get(outer, "inner.sym_arg0")

    if batch_outer_0 == () and batch_outer_1 == ():
        outer.print()
        pushed = ModulePusher().push_down_modules(outer, traversal=new_traversal(end_depth=3))
        pushed.print()

        # TODO: maybe extract should elim not input modules itself?
        extracted = extract_symbols_get(pushed, {"outer.sym_arg0", "outer.sym_arg1"})
        undo_push = extracted.update(
            rc.restrict(Matcher.regex("^outer( |$)"), start_depth=1),
            lambda x: x.rename("add_inner" if separate_mod else "inner.call"),
        )
        undo_push.print()
        assert undo_push == outer

    push = ModulePusher().push_down_modules(outer, traversal=new_traversal(), skip_module={"inner.call"})
    push.print()
    assert len(push.get(Module)) == 1
    assert {s.symbol.name for s in push.get_unique(Module).cast_module().spec.arg_specs} == {
        s.symbol.name for s in inner_call.spec.arg_specs
    }
    assert len(push.get(Symbol)) == 2
    torch.testing.assert_close(push.evaluate(), outer.evaluate())


@pytest.mark.parametrize("separate_mod", [False, True])
def test_complex_push_down(separate_mod: bool):
    torch.manual_seed(238)

    r = partial(run_test_push_down_mod, separate_mod=separate_mod)

    r()
    r(batch_inner_0=(3,))
    r(batch_inner_0=(3, 7))
    r(batch_inner_1=(3,))
    r(batch_inner_1=(3, 7))
    r(batch_inner_0=(3,), batch_inner_1=(3,))
    r(batch_inner_0=(7,), batch_inner_1=(7,))
    r(batch_inner_0=(3,), batch_inner_1=(7, 3))
    r(batch_inner_0=(7, 3), batch_inner_1=(7, 3))
    r(batch_inner_0=(7, 3), batch_inner_1=(3,))
    r(batch_inner_0=(1, 7, 3), batch_inner_1=(3,))

    r(batch_outer_0=(3,))
    r(batch_outer_0=(3, 7))
    r(batch_outer_1=(3,))
    r(batch_outer_1=(3, 7))
    r(batch_outer_0=(3,), batch_outer_1=(3,))
    r(batch_outer_0=(7,), batch_outer_1=(7,))
    r(batch_outer_0=(3,), batch_outer_1=(7, 3))
    r(batch_outer_0=(7, 3), batch_outer_1=(7, 3))
    r(batch_outer_0=(7, 3), batch_outer_1=(3,))
    r(batch_outer_0=(1, 7, 3), batch_outer_1=(3,))

    r(batch_inner_0=(3,), batch_outer_0=(3,))
    r(batch_inner_0=(3,), batch_outer_0=(4,))
    r(batch_inner_0=(3, 1), batch_outer_0=(4,), batch_outer_1=(6, 4))
    r(batch_inner_0=(3, 2), batch_inner_1=(2,), batch_outer_0=(4,), batch_outer_1=(6, 4))
    r(batch_inner_1=(7,), batch_outer_0=(4,), batch_outer_1=(6, 4))
    r(batch_inner_1=(3,), batch_outer_0=(4,))
    r(batch_inner_1=(3,), batch_outer_1=(4,))
    r(batch_inner_0=(3,), batch_outer_1=(4,))
    r(batch_inner_0=(3,), batch_outer_1=(6, 4))


def test_extract_traversal():
    s = """
    'add' Add
      'a' Module
        'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
        'b.arg' [] Scalar 73.3874 ! 'sym'
      'b' Module
        'sym'
        'a.arg' [] Scalar 28.3877 ! 'sym'
    """
    circ = P(s)
    out = extract_symbols_get(circ, "sym", traversal=new_traversal(term_early_at="a"))
    assert circ.get_unique("a") == out.get_unique("a")
    with pytest.raises(ExtractSymbolsBoundInputInconsistentError):
        extract_symbols_get(circ, "sym")


def test_push_down_module_nested_update_arg():
    s = """
    'a' Module
      'b' Module
        'c' Module
          'add' Add
            'sym_a' [] Symbol 05c93da1-ac33-47eb-b215-d1a8e8438a02
            'sym_b' [] Symbol 5ae54c16-b570-4abd-ad37-f527c25494c1
            'sym_c' [] Symbol cb22836a-b5cb-4e65-969e-563958249e81
          'arg_c' [] Scalar 3 ! 'sym_c'
        'arg_b' Einsum ij,j->i ! 'sym_b'
          'within_arg' Module
            'add_within' Add
              'sym_d' [] Symbol fbb99b36-0809-4eae-ae10-7d38a90d1386
              'w' [3] Scalar 4
            'k' [7] Scalar 7 ! 'sym_d'
          'x' [3] Scalar 8
      'arg_a' [] Scalar 82.83 ! 'sym_a'
    """
    circ = P(s)
    circ.print()

    pushed = ModulePusher().push_down_modules(circ, traversal=new_traversal())
    torch.testing.assert_close(pushed.evaluate(), circ.evaluate())

    pushed = ModulePusher(flatten_modules=False).push_down_modules(circ, traversal=new_traversal())
    torch.testing.assert_close(pushed.evaluate(), circ.evaluate())

    pushed = ModulePusher(flatten_modules=False).push_down_modules(
        circ, traversal=new_traversal(term_early_at={"sym_a", "sym_d"})
    )
    torch.testing.assert_close(pushed.evaluate(), circ.evaluate())

    pushed = ModulePusher(flatten_modules=False).push_down_modules(
        circ, traversal=new_traversal(term_early_at={"sym_a", "sym_d", "add"}), skip_module={"a", "c"}
    )
    assert isinstance(pushed.get_unique("within_arg"), Add)
    pushed.print()
    torch.testing.assert_close(pushed.evaluate(), circ.evaluate())

    pushed = ModulePusher(flatten_modules=True).push_down_modules(
        circ, traversal=new_traversal(term_early_at={"sym_a", "sym_d", "add"}), skip_module={"a", "c"}
    )
    assert isinstance(pushed.get_unique("within_arg"), Add)
    pushed.print()
    torch.testing.assert_close(pushed.evaluate(), circ.evaluate())


def test_get_free_symbols():
    s = """
    'a' [] Symbol 26051952-3d72-44dd-a5b0-3aa9ac49df60
    'a.bind' Module
      'a'
    'a.bind_unused' Module
      'a'
      'b' [] Symbol afa399cc-8db1-4670-9503-2aed00a1690d ! 'c' [] Symbol 706e3b9b-75d9-438f-b590-a7ee9c244104
    'a.bind_used' Module
      'a'
      'b' ! 'a'
    'b'
    """

    a, a_bind, a_bind_unused, a_bind_used, b = Parser(module_check_all_inputs_used=False).parse_circuits(s)
    assert get_free_symbols(a) == [a]
    assert get_free_symbols(a_bind) == [a]
    assert get_free_symbols(a_bind_unused) == [a, b]
    assert get_free_symbols(a_bind_used) == [b]


def test_complex_nesting_case():
    s = """
    'a' Module
      'b' Module
        'wrap' Add
          'inner_a' Module
            'inner_b' Module
              'add' Add
                'b_sym' [] Symbol d8d42e00-53bb-44df-8fe5-ce833805cc3e
                'b_other' [] Symbol 49ca8b31-7c1b-468c-a33e-c6a07a31e224
              'my_nested' Einsum -> ! 'b_sym'
                'a_sym' [] Symbol 43161ffc-dcf1-4163-a08b-54c6174472d3
            'a_val'  [4, 1, 1] Array rand ! 'a_sym'
          'b_sym'
        'b_other_val' [2] Array rand ! 'b_other'
        'my_nested' ! 'b_sym'
      'a_val2' [3, 1] Array rand ! 'a_sym'
    """
    c = cast_circuit(Parser()(s), TorchDeviceDtypeOp(dtype=torch.float64))

    mods = ["a", "b", "inner_a", "inner_b"]
    for perm in list(itertools.permutations(mods)):
        new = c
        for n in perm:
            # might miss some mods, whatever
            new = new.update(Matcher.regex(rf"^{n}(( )|$)") & Module, lambda x: x.cast_module().substitute())
            assert not any("internal_expand" in x.name for x in all_children(new))
        torch.testing.assert_close(new.evaluate(), c.evaluate())


def test_override_batch_sub_regression():
    s = """
    'b' Module
      'wrap' Add
        'inner_a' Module
          'add' Add
            'b_sym' [] Symbol d8d42e00-53bb-44df-8fe5-ce833805cc3e
            'b_other' [] Symbol 49ca8b31-7c1b-468c-a33e-c6a07a31e224
          'a_val' [4,1,1] Array rand ! 'b_sym'
        'b_sym'
      'b_other_val' [2] Array rand ! 'b_other'
      'a_val2' [] Scalar 23874 ! 'b_sym'
    """
    c = cast_circuit(Parser()(s), TorchDeviceDtypeOp(dtype=torch.float64))
    torch.testing.assert_close(c.cast_module().substitute().evaluate(), c.evaluate())
    assert not any("internal_expand" in x.name for x in all_children(c.cast_module().substitute()))

    s = """
    'b' Module
      'wrap' Add
        'inner_a' Module
          'add' Add
            'b_sym' [] Symbol d8d42e00-53bb-44df-8fe5-ce833805cc3e
            'b_other' [] Symbol 49ca8b31-7c1b-468c-a33e-c6a07a31e224
          'a_val' [4,1,1] Array rand ! 'b_sym'
        'b_sym'
      'b_other_val' [2] Array rand ! 'b_other'
      'a_val2' [3, 2] Scalar 23874 ! 'b_sym'
    """
    c = cast_circuit(Parser()(s), TorchDeviceDtypeOp(dtype=torch.float64))
    torch.testing.assert_close(c.cast_module().substitute().evaluate(), c.evaluate())
    print(c.cast_module().substitute())
    assert not any("internal_expand" in x.name for x in all_children(c.cast_module().substitute()))


def test_module_pusher_naming_edge_cases():
    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'b' Module
      'random_sym'
    """
    sym, c = Parser().parse_circuits(s)
    assert ModulePusher().push_down_modules(c, traversal=new_traversal()) == sym

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'a' Module
      'b' Module
        'random_sym'
    """
    sym, c = Parser().parse_circuits(s)
    assert ModulePusher().push_down_modules(c, traversal=new_traversal()) == sym

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'add' Add
      'random_sym'
      'random_sym'
    'b' Module
      'add'
    """
    sym, add, c = Parser().parse_circuits(s)
    assert ModulePusher().push_down_modules(c, traversal=new_traversal()) == add

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'add' Add
      'random_sym'
      'random_sym'
    'a' Module
      'b' Module
        'add'
    """
    sym, add, c = Parser().parse_circuits(s)
    assert ModulePusher().push_down_modules(c, traversal=new_traversal()) == add

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'add' Add
      'random_sym'
      'random_sym'
    'a' Module
      'b' Module
        'c' Module
          'add'
          'garbage' [] Scalar 3 ! 'unused0' [] Symbol 26f1c039-5491-458e-a908-1a90e258c806
        'garbage' ! 'unused0'
      'garbage' ! 'unused0'
    """
    sym, add, c = Parser(module_check_all_inputs_used=False).parse_circuits(s)
    assert ModulePusher().push_down_modules(c, traversal=new_traversal()) == add

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'add' Add
      'random_sym'
      'random_sym'
    'c' Module
      'add'
      'garbage' [] Scalar 3 ! 'unused0' [] Symbol 26f1c039-5491-458e-a908-1a90e258c806

    'a' Module
      'b' Module
        'c'
        'garbage' ! 'unused0'
      'garbage' ! 'unused0'
    """
    sym, add, c_mod, c = Parser(module_check_all_inputs_used=False).parse_circuits(s)
    assert ModulePusher().push_down_modules(c, traversal=new_traversal(term_early_at="c")) == c_mod

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'my_arg' [] Scalar 3874
    'b' Module
      'random_sym'
      'my_arg' ! 'random_sym'
    """
    sym, arg, c = Parser().parse_circuits(s)
    assert ModulePusher().push_down_modules(c, traversal=new_traversal()) == arg

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'my_arg' [] Scalar 3874
    'a' Module
      'b' Module
        'random_sym'
        'my_arg' ! 'random_sym'
    """
    sym, arg, c = Parser().parse_circuits(s)
    assert ModulePusher().push_down_modules(c, traversal=new_traversal()) == arg

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    0 'b' Add
      'add bind:b' Add
        'my_arg' [] Scalar 3874
        'my_arg' [] Scalar 3874
    1 'b' Module
      'outer' Add
        'add' Add
          'random_sym'
          'random_sym'
      'my_arg' ! 'random_sym'
    """
    sym, new_add, c = Parser().parse_circuits(s)

    def namer(*args):
        print(args)
        return default_nested_module_namer()(*args)

    assert ModulePusher(namer=namer).push_down_modules(c, traversal=new_traversal()) == new_add


def test_module_sub():
    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'b' Module
      'random_sym'
    """
    sym, c = Parser().parse_circuits(s)
    assert c.cast_module().substitute() == sym

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'a' Module
      'b' Module
        'random_sym'
    """
    sym, c = Parser().parse_circuits(s)
    assert substitute_all_modules(c) == sym
    assert substitute_all_modules(c.cast_module().substitute()) == sym

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'add' Add
      'random_sym'
      'random_sym'
    'b' Module
      'add'
    """
    sym, add, c = Parser().parse_circuits(s)
    assert c.cast_module().substitute() == add

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'add' Add
      'random_sym'
      'random_sym'
    'a' Module
      'b' Module
        'add'
    """
    sym, add, c = Parser().parse_circuits(s)
    assert substitute_all_modules(c) == add
    assert substitute_all_modules(c.cast_module().substitute()) == add

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'add' Add
      'random_sym'
      'random_sym'
    'a' Module
      'b' Module
        'c' Module
          'add'
          'garbage' [] Scalar 3 ! 'unused0' [] Symbol 26f1c039-5491-458e-a908-1a90e258c806
        'garbage' ! 'unused0'
      'garbage' ! 'unused0'
    """
    sym, add, c = Parser(module_check_all_inputs_used=False).parse_circuits(s)
    assert substitute_all_modules(c) == add
    assert substitute_all_modules(c.cast_module().substitute()) == add

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'my_arg' [] Scalar 3874
    'b' Module
      'random_sym'
      'my_arg' ! 'random_sym'
    """
    sym, arg, c = Parser().parse_circuits(s)
    assert c.cast_module().substitute() == arg

    s = """
    'random_sym' [] Symbol ded02d73-6102-4f7f-a8cf-15b1065e226e
    'my_arg' [] Scalar 3874
    'a' Module
      'b' Module
        'random_sym'
        'my_arg' ! 'random_sym'
    """
    sym, arg, c = Parser().parse_circuits(s)
    assert substitute_all_modules(c.cast_module().substitute()) == arg
    assert substitute_all_modules(c) == arg



import pytest

from rust_circuit import *


def test_schedule_serialize():
    circ = Einsum.from_einsum_string("ab,bc->ac", Array.randn(2, 3), Array.randn(3, 4))
    schedule = optimize_to_schedule(circ)
    serialized = schedule.serialize()
    print(serialized)
    deserialized = Schedule.deserialize(serialized)
    print(deserialized)


@pytest.mark.skip(reason="we don't have server test set up yet")
def test_remote_execution():
    circ = Einsum.from_einsum_string("ab,bc->ac", Array.randn(2, 3), Array.randn(3, 4))
    schedule = optimize_to_schedule(circ)
    result = schedule.evaluate_remote("http://0.0.0.0:9876")
    print(result)


if __name__ == "__main__":
    # test_schedule_serialize()
    test_remote_execution()



import rust_circuit as rc
from rust_circuit import Parser
from rust_circuit.ui.ui import circuit_graph_ui

parent = Parser()(
    """0 Add
  4 'asdasd' Add
    1 'asdf' Add
      2 [1] Add
        6 'Z' [1] Scalar 12
  5 'ashudashk' Add
    3 'whe' Add
      2
  7 'hi' [11] Scalar 27
  8 'blah' [?] Symbol"""
)

parent = Parser()(
    """'b' Add # line length: 4 # node count: 67
  'b.m' Module # line length: 7 # node count: 66
    'm.ln_call' Module # line length: 7 # node count: 28
      'm' Einsum h,oh->o # line length: 15 # node count: 8
        'm.act' GeneralFunction gelu # line length: 21 # node count: 6
          'm.pre' Add # line length: 4 # node count: 5
            'm.pre_mul' Einsum i,hi->h # line length: 15 # node count: 3
              'm.input' [?] Symbol 13ec4cdc-5f25-4969-870b-0bfa2300187b # line length: 48 # node count: 1
              'm.w.proj_in' [?,?] Symbol 5217f963-0cdb-460e-bb1f-f82f7fbb3cd9 # line length: 50 # node count: 1
            'm.w.in_bias' [?] Symbol c870ec00-8c6f-4080-907c-703ea85dde48 # line length: 48 # node count: 1
        'm.w.proj_out' [?,?] Symbol fdefa9af-a7d6-4a38-a7ed-5ce816c6efe7 # line length: 50 # node count: 1
      'm.ln' Module ! 'm.input' # line length: 7 # node count: 19
        'ln' Add # line length: 4 # node count: 15
          'ln.w.bias' [?] Symbol 621c7792-0177-45ab-87c5-7ff1c3bec487 # line length: 48 # node count: 1
          'ln.y_scaled' Einsum h,h->h # line length: 14 # node count: 13
            'ln.y' Einsum h,->h # line length: 13 # node count: 11
              'ln.mean_subbed' Add # line length: 4 # node count: 6
                'ln.input' [?] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c # line length: 48 # node count: 1
                'ln.neg_mean' Einsum h,z,->z # line length: 15 # node count: 5
                  'ln.input' # line length: 48 # node count: 1
                  'ln.neg' [1] Scalar -1 # line length: 14 # node count: 1
                  'ln.c.recip_hidden_size' GeneralFunction reciprocal # line length: 27 # node count: 3
                    'ln.c.hidden_size' GeneralFunction last_dim_size # line length: 30 # node count: 2
                      'ln.input' # line length: 48 # node count: 1
              'ln.rsqrt' GeneralFunction rsqrt # line length: 22 # node count: 10
                'ln.var_p_eps' Add # line length: 4 # node count: 9
                  'ln.c.eps' [] Scalar 0.00001 # line length: 18 # node count: 1
                  'ln.var' Einsum h,h,-> # line length: 14 # node count: 7
                    'ln.mean_subbed' # line length: 4 # node count: 6
                    'ln.mean_subbed' # line length: 4 # node count: 6
                    'ln.c.recip_hidden_size' # line length: 27 # node count: 3
            'ln.w.scale' [?] Symbol 0fa341c3-34b3-4699-847f-08674808b28a # line length: 48 # node count: 1
        'm.ln.input' [?,?] Symbol a0fe2ee1-77bc-4afd-bf98-3b34212d944b ! 'ln.input' # line length: 50 # node count: 1
        'm.ln.w.bias' [?] Symbol 049d1137-70a9-495b-b0d1-01625ec05540 ! 'ln.w.bias' # line length: 48 # node count: 1
        'm.ln.w.scale' [?] Symbol b325c75f-f8b8-41c5-81d9-4ed21dc82208 ! 'ln.w.scale' # line length: 48 # node count: 1
    'b.m.input' Add ! 'm.ln.input' # line length: 4 # node count: 52
      'b.a' Module # line length: 7 # node count: 51
        'a.ln_call' Module # line length: 7 # node count: 49
          'a.on_inp' Module # line length: 7 # node count: 29
            'a' Einsum shV,hdV->sd # line length: 19 # node count: 27
              'a.comb_v' Einsum hqk,hkV->qhV # line length: 20 # node count: 25
                'a.attn_probs' GeneralFunction softmax # line length: 24 # node count: 21
                  'a.attn_scores' Add # line length: 4 # node count: 20
                    'a.attn_scores_raw' Einsum hqc,hkc,,qk->hqk # line length: 24 # node count: 11
                      'a.q' Einsum qd,hcd->hqc # line length: 19 # node count: 3
                        'a.q.input' [?,?] Symbol 4f80d1a1-86a4-4e44-94f7-909ec7089061 # line length: 50 # node count: 1
                        'a.w.q' [?,?,?] Symbol 665efa60-d86c-40d5-92b2-b96d11686a8b # line length: 52 # node count: 1
                      'a.k' Einsum kd,hcd->hkc # line length: 19 # node count: 3
                        'a.k.input' [?,?] Symbol 664bddee-28ca-47e7-9fb7-9a718de06619 # line length: 50 # node count: 1
                        'a.w.k' [?,?,?] Symbol 41177709-446d-4588-b9e5-c2bbf59d53a0 # line length: 52 # node count: 1
                      'a.c.div_head_size' GeneralFunction rsqrt # line length: 22 # node count: 4
                        'a.c.head_size' GeneralFunction last_dim_size # line length: 30 # node count: 3
                          'a.c.bias_for_head_size' Einsum ijk->j # line length: 14 # node count: 2
                            'a.w.k' # line length: 52 # node count: 1
                      'a.mask' [?,?] Symbol ccfe5bc9-b402-42dd-a5e1-191e6fb7c268 # line length: 50 # node count: 1
                    'a.score_neg_inf_bias' Einsum qk,o->oqk # line length: 17 # node count: 9
                      'a.not_mask' Module # line length: 7 # node count: 7
                        'not_mask' Add # line length: 4 # node count: 5
                          'one' [] Scalar 1 # line length: 12 # node count: 1
                          'not_mask.neg_mask' Einsum ,-> # line length: 11 # node count: 3
                            'not_mask.input' [] Symbol b46f6370-11e1-4535-aabc-94554c234673 # line length: 47 # node count: 1
                            'neg_one' [] Scalar -1 # line length: 13 # node count: 1
                        'a.mask' ! 'not_mask.input' # line length: 50 # node count: 1
                      'a.neg_inf' [1] Scalar -10000 # line length: 18 # node count: 1
                'a.v' Einsum kd,hVd->hkV # line length: 19 # node count: 3
                  'a.v.input' [?,?] Symbol 8fd4c632-7f28-49ee-84cc-3dde997e0693 # line length: 50 # node count: 1
                  'a.w.v' [?,?,?] Symbol 79b6ebff-f9d0-411a-bcdc-530cc13e1524 # line length: 52 # node count: 1
              'a.w.o' [?,?,?] Symbol 11a116cb-2168-4725-a06f-1b61a8ca6797 # line length: 52 # node count: 1
            'a.input' [?,?] Symbol f9eabd07-e2ab-4ed4-8b4a-c9c039d61835 ! 'a.q.input' # line length: 50 # node count: 1
            'a.input' ! 'a.k.input' # line length: 50 # node count: 1
            'a.input' ! 'a.v.input' # line length: 50 # node count: 1
          'a.ln' Module ! 'a.input' # line length: 7 # node count: 19
            'ln' # line length: 4 # node count: 15
            'a.ln.input' [?,?] Symbol 85fb9501-ce13-4aa5-ab98-558020a3daec ! 'ln.input' # line length: 50 # node count: 1
            'a.ln.w.bias' [?] Symbol cecc5585-abab-461e-afcd-664a1dd80037 ! 'ln.w.bias' # line length: 48 # node count: 1
            'a.ln.w.scale' [?] Symbol e8f6c84d-9593-4383-8f5d-76b2ea589c8e ! 'ln.w.scale' # line length: 48 # node count: 1
        'b.input' [?,?] Symbol 5837c4fd-f5ac-4bff-8456-abf3e95bcf36 ! 'a.ln.input' # line length: 50 # node count: 1
      'b.input' # line length: 50 # node count: 1
  'b.m.input' # line length: 4 # node count: 52"""
)

circuit_graph_ui(
    parent,
    annotators={"test": lambda circ: circ.name, "children": lambda circ: str(circ.num_children)},
    enabled_annotators=["children"],
    # default_hidden=parent.get(ui_default_hidden_matcher),
    default_shown=parent.get(rc.Matcher(rc.Einsum, rc.Add, rc.Module)),
)



import uuid
from functools import lru_cache
from typing import *

import torch

import rust_circuit as rc

# based on https://github.com/huggingface/transformers/blob/6767ce71d661ae57afa38b6cd0d533e3dfd463fa/src/transformers/models/codegen/modeling_codegen.py
# note: different models do rotary pos emb somewhat differently

# NOTE: probably faster to just fuse this with the computation rather than cache
@lru_cache(maxsize=2)
def _rotary_embs(dim, seq_len, device):
    inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2) / dim))
    freqs = torch.einsum("i , j -> i j", torch.arange(seq_len, dtype=torch.float), inv_freq).to(device).float()
    sin = duplicate_interleave(freqs.sin())[None, :, None, :]
    cos = duplicate_interleave(freqs.cos())[None, :, None, :]
    return sin, cos


def rotate_every_two(x):
    x1 = x[:, :, :, ::2]
    x2 = x[:, :, :, 1::2]
    x = torch.stack((-x2, x1), dim=-1)
    return x.flatten(-2)  # in einsum notation: rearrange(x, '... d j -> ... (d j)')


def duplicate_interleave(m):
    """
    A simple version of `torch.repeat_interleave` for duplicating a matrix while interleaving the copy.
    """
    dim0 = m.shape[0]
    m = m.view(-1, 1)  # flatten the matrix
    m = m.repeat(1, 2)  # repeat all elements into the 2nd dimension
    m = m.view(dim0, -1)  # reshape into a matrix, interleaving the copy
    return m


class ApplyRotaryPosEmb(rc.GeneralFunctionSpecBase):
    @property
    def path(self) -> str:
        return "rust_circuit.generalfuncs.rotary_pos_emb:ApplyRotaryPosEmb"

    def compute_hash_bytes(self) -> bytes:
        return uuid.UUID("a00070be-b82b-495f-953f-633d0ab6829b").bytes

    def function(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        """x can be either q or k"""
        seq_len, two, halfdim = x.shape[-3:]
        assert two == 2
        x = x.view(*x.shape[:-2], 1, -1)
        sin, cos = _rotary_embs(2 * halfdim, seq_len, device=x.device)
        x = (x * cos) + (rotate_every_two(x) * sin)
        x = x.view(*x.shape[:-2], 2, -1)
        return x

    def get_shape_info(self, x_shape: rc.Shape) -> rc.GeneralFunctionShapeInfo:  # type: ignore[override]
        if len(x_shape) < 3:
            raise ValueError(f"x must have at least 3 dimensions, but got {len(x_shape)}")
        if x_shape[-2] != 2:
            raise ValueError(f"Second to last dimension of x must be 2, but got {x_shape[-2]}")
        return rc.GeneralFunctionShapeInfo(
            x_shape,
            num_non_batchable_output_dims=2,
            input_batchability=[True],
        )



from typing import Dict, Literal, Optional

import einops
import jax.numpy as jnp
import numpy as np
import torch

from interp.model.blocks import BatchNormFixed, LayerNorm
from interp.model.gpt_model import Gpt
from interp.model.gpt_modules import Attention, GptBlock, PosEncType, id_norm
from rust_circuit import module_library as mod_l

from . import _rust as rc


def from_converted(arr: jnp.ndarray, name: Optional[str] = None):
    return rc.Array(torch.from_numpy(np.array(arr)), name=name)


def from_converted_tup(arr: jnp.ndarray, s: str):
    return (s, from_converted(arr))


def get_norm(norm_b):
    if not isinstance(norm_b, (LayerNorm, BatchNormFixed)):
        raise NotImplementedError("unsupported norm type, currently only supports ln/bn fixed", type(norm_b))
    if norm_b.epsilon != 1e-5:
        raise NotImplementedError("TODO: support variable epsilon")
    nt = "ln" if isinstance(norm_b, LayerNorm) else "bn"
    out = [
        from_converted_tup(norm_b.variables["params"]["bias"], s=f"{nt}.w.bias"),
        from_converted_tup(norm_b.variables["params"]["scale"], s=f"{nt}.w.scale"),
    ]
    if nt == "bn":
        out.extend(
            [
                from_converted_tup(norm_b.variables["params"]["mean"], s=f"{nt}.mean"),
                from_converted_tup(norm_b.variables["params"]["var"], s=f"{nt}.var"),
            ]
        )

    return out, nt


def get_attention(attention_b: Attention, use_pos: bool):
    if attention_b.softmax_type != "softmax":
        raise NotImplementedError
    q_w, k_w, v_w = attention_b.get_qkv_mats()
    o_w = einops.rearrange(
        attention_b.project_output.get_weights(),
        "hidden_out (num_heads head_size) -> num_heads hidden_out head_size",
        num_heads=attention_b.num_heads,
    )

    out = [
        from_converted_tup(q_w, "a.w.q"),
        from_converted_tup(k_w, "a.w.k"),
        from_converted_tup(v_w, "a.w.v"),
        from_converted_tup(o_w, "a.w.o"),
    ]

    if attention_b.bias:
        q_bias, k_bias, v_bias = einops.rearrange(
            attention_b.attn_weights.get_bias(),
            "(k num_heads head_size) -> k num_heads head_size",
            k=3,
            num_heads=q_w.shape[0],
        )
        o_bias = attention_b.project_output.get_bias()
        out.extend(
            [
                from_converted_tup(q_bias * (2 if use_pos else 1), "a.w.q_bias"),
                from_converted_tup(k_bias * (2 if use_pos else 1), "a.w.k_bias"),
                from_converted_tup(v_bias, "a.w.v_bias"),
                from_converted_tup(o_bias, "a.w.o_bias"),
            ]
        )

    return out


def get_mlp(block_b: GptBlock):
    proj_in = block_b.linear1.get_weights()
    in_bias = block_b.linear1.get_bias()

    proj_out = block_b.linear2.get_weights()

    if block_b.mlp_act_type == "bilinear":
        # only half of params are used in this case
        proj_out = einops.rearrange(proj_out, "hidden_out (a mlp_proj) -> a hidden_out mlp_proj", a=2)[0]

    out = [
        from_converted_tup(proj_in, "m.w.proj_in"),
        from_converted_tup(in_bias, "m.w.in_bias"),
        from_converted_tup(proj_out, "m.w.proj_out"),
    ]

    if block_b.linear2.use_bias:
        out.append(from_converted_tup(block_b.linear2.get_bias(), "m.w.out_bias"))

    return out


def get_block(block_b: GptBlock, pos_enc_type: PosEncType):
    all_inputs: dict[str, rc.Circuit] = {}

    use_norm = block_b.norm1 is not id_norm

    attn_pos = pos_enc_type == "shortformer"

    norm_type = None
    if use_norm:
        norm_circs_attn, norm_type = get_norm(block_b.norm1)
        all_inputs.update(mod_l.apply_prefix(dict(norm_circs_attn), "a"))

    all_inputs.update(get_attention(block_b.attention, use_pos=attn_pos))

    if block_b.use_mlp:
        if use_norm:
            mlp_norm_circs, norm_type_m = get_norm(block_b.norm2)
            assert norm_type_m == norm_type
            all_inputs.update(mod_l.apply_prefix(dict(mlp_norm_circs), "m"))

        all_inputs.update(get_mlp(block_b))

    return all_inputs


def get_model(model_b: Gpt):
    all_inputs: Dict[str, rc.Circuit] = {}
    for block_i in range(model_b.num_layers):
        block_b = model_b.blocks[block_i]
        all_inputs.update(
            {mod_l.add_number(s, block_i): c for s, c in get_block(block_b, pos_enc_type=model_b.pos_enc_type).items()}
        )

    if model_b.norm_type != "none" and model_b.use_norm_output:
        norm, _ = get_norm(model_b.norm_output)
        all_inputs.update(mod_l.apply_prefix(dict(norm), "final"))

    unembed_name = "t.w.unembed"
    if model_b.classifier:
        unembedding = from_converted(model_b.embedding.linear_out.get_weights(), name=unembed_name)
    else:
        unembedding = from_converted(model_b.embedding.token_unembedding.embedding, name=unembed_name)
    all_inputs[unembedding.name] = unembedding

    if model_b.classifier:
        output_bias = from_converted(model_b.embedding.linear_out.get_bias(), name="t.w.unembed_bias")
        all_inputs[output_bias.name] = output_bias

    return all_inputs


def get_model_info(model_b: Gpt, model_class: str = "GPTBeginEndToks"):
    norm_type: Optional[Literal["ln", "bn"]]
    if model_b.norm_type == "none":
        norm_type = None
    elif model_b.norm_type == "layer_norm":
        norm_type = "ln"
    elif model_b.norm_type == "batch_norm_fixed":
        norm_type = "bn"
    else:
        raise NotImplementedError("unsupported norm type, currently only supports ln/bn fixed", model_b.norm_type)
    return mod_l.TransformerInfo(
        mod_l.TransformerParams(
            mod_l.TransformerBlockParams(
                norm_type=norm_type,
                attn_bias=model_b.attn_bias,
                attn_pos=model_b.pos_enc_type == "shortformer",
                use_mlp=model_b.use_mlp,
                mlp_act_type=model_b.mlp_act_type,
                mlp_output_bias=model_b.mlp_bias,
            ),
            num_layers=model_b.num_layers,
            use_norm_output=model_b.use_norm_output,
            output_bias=model_b.classifier,
        ),
        model_class=model_class,
        pos_enc_type=model_b.pos_enc_type,
        causal_mask=model_b.causal_mask,
    )


def get_bound_model(model_b: Gpt, model_class: str = "GPTBeginEndToks"):
    all_inputs = mod_l.rename_circs_to_keys(get_model(model_b), "_arr")
    tok_embeds = from_converted(model_b.embedding.token_embedding.embedding, name="t.w.tok_embeds")
    pos_embeds = from_converted(model_b.embedding.position_embedding.embedding, name="t.w.pos_embeds")
    info = get_model_info(model_b, model_class=model_class)
    model_ret = info.params.get()
    circ = rc.module_new_bind(model_ret.body, *list(all_inputs.items()), name="t.bind_w")

    return circ, (tok_embeds, pos_embeds), info, model_ret



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Sequence, Tuple, Optional, Union

from dataclasses import dataclass

import torch

from rust_circuit.index_util.internal.resolve import IntLitIdx, OutAxisIdx

if TYPE_CHECKING:
    ViewIdxExpr = Union[OutAxisIdx, IntLitIdx]

#
# Planning
#


@dataclass
class ConstantIndexPlan:
    constants: list[Optional[int]]


def plan_constant_index(indices: Sequence[ViewIdxExpr]) -> Tuple[ConstantIndexPlan, list[OutAxisIdx]]:
    constants: list[Optional[int]] = []
    new_indices: list[OutAxisIdx] = []
    for idx in indices:
        if isinstance(idx, IntLitIdx):
            constants.append(idx.value)
        else:
            constants.append(None)
            new_indices.append(idx)
    return ConstantIndexPlan(constants), new_indices


@dataclass
class DiagonalPlan:
    """
    Each tuple in 'diagonals' has the semantics of replacing the first axis in the tuple with the diagonal along both
    axes.

    Note that this is not the same as the semantics of 'torch.diagonal', which places the diagonal at the end.

    We implement the semantics here in terms of PyTorch's primitives by using 'torch.diagonal' to get the diagonal, and
    then using 'torch.transpose' to move the diagonal to the position of the first axis.
    """

    diagonals: list[Tuple[int, int]]


def plan_diagonals(indices: Sequence[OutAxisIdx]) -> Tuple[DiagonalPlan, list[OutAxisIdx]]:
    diagonals: list[Tuple[int, int]] = []
    first_seen: dict[int, int] = {}
    new_indices: list[OutAxisIdx] = []
    for i, idx in enumerate(indices):
        prev = first_seen.get(idx.axis)
        if prev is not None:
            diagonals.append((prev, i))
        else:
            first_seen[idx.axis] = i
            new_indices.append(idx)
    return DiagonalPlan(diagonals), new_indices


@dataclass
class TransposePlan:
    perm: list[int]


def plan_transpose(indices: Sequence[OutAxisIdx]) -> Tuple[TransposePlan, list[OutAxisIdx]]:
    perm = sorted(range(len(indices)), key=lambda i: indices[i].axis)
    return TransposePlan(perm), [indices[i] for i in perm]


@dataclass
class BroadcastPlan:
    """
    If 'broadcast_axes[i]' is not 'None', then a new axis should be inserted at position 'i', and the new axis should
    be broadcast to the same size as the axis 'broadcast_axes[i]' in the output.
    """

    broadcast_axes: list[Optional[int]]


def plan_broadcast(out_rank: int, indices: Sequence[OutAxisIdx]) -> Tuple[BroadcastPlan, list[OutAxisIdx]]:
    broadcast_axes: list[Optional[int]] = []
    new_indices: list[OutAxisIdx] = []
    j = 0
    for i in range(out_rank):
        while j < len(indices):
            idx = indices[j]
            if idx.axis == i:
                broadcast_axes.append(None)
                new_indices.append(idx)
                j += 1
                break
            else:
                assert idx.axis > i, "indices must be sorted by prior transpose step"
                broadcast_axes.append(i)
                new_indices.append(OutAxisIdx(i))
                break
        else:
            broadcast_axes.append(i)
            new_indices.append(OutAxisIdx(i))

    return BroadcastPlan(broadcast_axes), new_indices


@dataclass
class ViewPlan:
    input_rank: int
    constant_index: ConstantIndexPlan
    diagonal: DiagonalPlan
    transpose: TransposePlan
    broadcast: BroadcastPlan


def plan_view(out_rank: int, indices: Sequence[ViewIdxExpr]) -> ViewPlan:
    """
    Given an indexing expression

        arr[e_0, e_1, ..., e_{k-1}]

    expressed in terms of iteration variables ("out axes")

        i_0, i_1, ..., i_{n-1}

    returns a "plan" representing a function 'f' such that

        f(arr)[i_0, i_1, ..., i_{n-1}] == arr[e_0, e_1, ..., e_{k-1}]
    """

    rank = len(indices)
    constant_index, indices = plan_constant_index(indices)
    diagonal, indices = plan_diagonals(indices)
    transpose, indices = plan_transpose(indices)
    broadcast, indices = plan_broadcast(out_rank, indices)
    assert indices == [OutAxisIdx(i) for i in range(out_rank)]
    return ViewPlan(rank, constant_index, diagonal, transpose, broadcast)


#
# Execution
#


def execute_plan(plan: ViewPlan, out_shape: Sequence[Optional[int]], arr: torch.Tensor) -> torch.Tensor:
    batch_rank = arr.ndim - plan.input_rank
    assert batch_rank >= 0

    # Constant index
    constant_indices = [slice(None)] * batch_rank + [
        slice(None) if value is None else value for value in plan.constant_index.constants
    ]
    arr = arr[constant_indices]

    # Diagonals
    for axis1, axis2 in plan.diagonal.diagonals:
        arr = arr.diagonal(dim1=batch_rank + axis1, dim2=batch_rank + axis2)
        new_rank = arr.ndim
        perm = list(range(new_rank - 1))
        perm.insert(batch_rank + axis1, new_rank - 1)
        arr = arr.permute(perm)

    # Transpose
    arr = arr.permute(list(range(batch_rank)) + [batch_rank + i for i in plan.transpose.perm])

    # Broadcast
    assert len(out_shape) == len(plan.broadcast.broadcast_axes), "out_shape should not have batch dimensions"
    full_out_shape: list[int] = []
    for i, axis in enumerate(plan.broadcast.broadcast_axes):
        if axis is None:
            size = out_shape[i]
            assert size is None or size == arr.shape[batch_rank + i]
            full_out_shape.append(arr.shape[batch_rank + i])
        else:
            arr = arr.unsqueeze(batch_rank + i)
            size = out_shape[i]
            assert size is not None
            full_out_shape.append(size)
    arr = arr.expand(list(arr.shape[: arr.ndim - len(full_out_shape)]) + full_out_shape)

    return arr



import uuid

import pytest

from rust_circuit import Add, Array, Cumulant, Einsum, Symbol, print_circuit_type_check


def test_circuit_type():
    assert print_circuit_type_check(Cumulant) is Cumulant
    assert isinstance(Symbol([2, 3], uuid.uuid4(), name="hi"), Symbol)
    assert print_circuit_type_check(Symbol) is Symbol
    assert print_circuit_type_check(Array) is Array
    assert print_circuit_type_check(Einsum) is Einsum
    assert print_circuit_type_check(Add) is Add
    with pytest.raises(TypeError):
        print_circuit_type_check(int)  # type:ignore
    with pytest.raises(TypeError):
        print_circuit_type_check(None)  # type:ignore
    assert Symbol([2, 3], uuid.uuid4(), name="hi").is_leaf()
    assert not Symbol([2, 3], uuid.uuid4(), name="hi").is_var()


if __name__ == "__main__":
    test_circuit_type()



import pandas

# wget http://www.naturalnumbers.org/P-100000.txt
df = pandas.read_csv("P-100000.txt")
arr = df[" 2"]
lst = list(arr[10_000 < arr][:5_000])
with open("primes.bin", "wb") as f:
    for x in lst:
        f.write(x.to_bytes(8, "little", signed=False))



from rust_circuit import *

P = Parser()


# correctness tests are in rea
def test_add_removable():
    circ = P(
        """
0 Add
  1 [4,1,2] Scalar 1
  2 'hi' [1,1,2] Symbol"""
    ).cast_add()
    print(add_elim_removable_axes_weak(circ))


def test_einsum_removable():
    circ = P(
        """
0 Einsum azc,yzc->azc
  1 [4,1,2] Scalar 1
  2 'hi' [1,1,2] Symbol"""
    ).cast_einsum()
    print(einsum_elim_removable_axes_weak(circ))


def test_einsum_permute_rearrange():
    circ = P(
        """
0 Einsum abc->bca
  2 'hi' [4,3,2] Symbol"""
    ).cast_einsum()
    print(einsum_permute_to_rearrange(circ))


def test_basic_simp():
    circ = P(
        """
0 Einsum azc,yzc->azc
  8 Rearrange a b c -> c b a
    1 [3,1,4] Scalar 1
  2 [1,1,3] Rearrange a b c -> a b c
    3 Concat 2
      4 'a' [1,1,1] Symbol
      5 'a' [1,1,1] Symbol
      10 Einsum abc->bca
        6 'b' [1,1,1] Symbol
    """
    )
    print(simp(circ))


if __name__ == "__main__":
    test_add_removable()
    test_einsum_removable()
    test_basic_simp()
    test_einsum_permute_rearrange()



from typing import Callable, Optional

import pytest
import torch

from rust_circuit import (
    Add,
    Array,
    Circuit,
    DiscreteVar,
    Einsum,
    Index,
    IterativeMatcher,
    PrintOptions,
    Regex,
    Scalar,
    TorchDeviceDtypeOp,
    cast_circuit,
    restrict,
)
from rust_circuit.causal_scrubbing.dataset import Dataset
from rust_circuit.causal_scrubbing.experiment import Experiment, ExperimentEvalSettings
from rust_circuit.causal_scrubbing.hypothesis import Correspondence, ExactSampler, FixedOtherSampler
from rust_circuit.causal_scrubbing.hypothesis import FuncSampler as FS
from rust_circuit.causal_scrubbing.hypothesis import (
    IllegalCorrespondenceError,
    InterpNode,
    UncondSampler,
    corr_root_matcher,
)
from rust_circuit.causal_scrubbing.testing_utils import IntDataset, loss_fn
from rust_circuit.py_utils import I

label_cond = FS(lambda d: d.labels.value)


def run_experiment(
    c: Circuit,
    ds: Dataset,
    corr: Correspondence,
    assert_loss_fn: Optional[Callable[[float], bool]] = None,
    num_examples=1,
    num_seeds=10,
    treeify=True,
    check=True,
):
    for seed in range(num_seeds):
        ex = Experiment(c, ds, corr, random_seed=seed, check=check)
        eval_settings = ExperimentEvalSettings(device_dtype="cpu")
        scrubbed = ex.scrub(treeify=treeify, ref_ds_or_num=num_examples)
        out = scrubbed.evaluate(eval_settings)
        if assert_loss_fn is not None:
            loss = loss_fn(scrubbed.ref_ds, out)
            assert assert_loss_fn(loss), loss


def datum_loss(c: Circuit, dataset: Dataset):
    datum = Array(dataset.xs.value[0], name="ref")
    c = IterativeMatcher("xs").update(c, lambda _: datum)
    out = cast_circuit(c, TorchDeviceDtypeOp(dtype="float32", device="cpu")).evaluate()
    return loss_fn(dataset[0], out)


def num_input_tags(c: Circuit):
    m = IterativeMatcher(Regex(r"\.*_scrub_\.*"))
    c.print()
    return len(m.get(c))


# === Corr checks ===
def test_corr_well_defined_matches_many():
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    second = Index(xs, I[1], name="x0")
    C = Add(first, second, name="C")

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    x0 = C_t.make_descendant(name="x0", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(x0, to_C.chain("x0"))

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])

    run_experiment(C, dataset, corr)


def test_corr_well_defined_matches_none():
    xs = Array(torch.tensor([0.0, 1.0]), name="not xs")
    first = Index(xs, I[0], name="x0")
    C = Add(first, Scalar(5), name="C")

    C_t = InterpNode(name="Ct", cond_sampler=UncondSampler())
    corr = Correspondence()
    corr.add(C_t, corr_root_matcher)

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])
    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)

    xs = Array(torch.tensor([0.0, 1.0]), name="not xs")
    first = Index(xs, I[0], name="x0")
    C = Add(first, Scalar(5), name="C")

    C_t = InterpNode(name="Ct", cond_sampler=UncondSampler())
    corr = Correspondence()
    corr.add(C_t, corr_root_matcher)
    x0 = C_t.make_descendant(name="x0", cond_sampler=UncondSampler())
    corr.add(x0, corr_root_matcher.chain("not_x0"))

    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)


def test_corr_injective_simple():
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    second = Index(xs, I[1], name="x1")
    C = Add(first, second, name="C")

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    x0_1 = C_t.make_descendant(name="x0_1", cond_sampler=UncondSampler())
    x0_2 = C_t.make_descendant(name="x0_2", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(x0_1, to_C.chain("x0"))
    corr.add(x0_2, to_C.chain("x0"))

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])
    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)


def test_corr_injective_subsumed():
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    A = Einsum.scalar_mul(first, 2)
    C = Add(first, A, name="C")

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    A = C_t.make_descendant(name="A", cond_sampler=UncondSampler())
    A_x0 = A.make_descendant(name="A_x0", cond_sampler=UncondSampler())
    x0 = C_t.make_descendant(name="x0", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A, to_C.chain("A"))
    corr.add(x0, to_C.chain("x0"))
    corr.add(A_x0, restrict(to_C.chain("x0"), term_early_at="A"))

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])
    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)


def test_corr_injective_checking_leaves_insufficient():
    # An example where the leaves of the interpretation map to different nodes in
    # the treeified model, but where intermediate interp nodes map to the same one!
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    second = Index(xs, I[1], name="x1")
    added = Add(first, second, name="x2")
    C = Add(added, name="C")

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    added_1 = C_t.make_descendant(name="x2_1", cond_sampler=UncondSampler())
    added_2 = C_t.make_descendant(name="x2_2", cond_sampler=UncondSampler())
    x0 = added_1.make_descendant(name="x0", cond_sampler=UncondSampler())
    x1 = added_2.make_descendant(name="x1", cond_sampler=UncondSampler())

    to_C = corr_root_matcher
    to_added_1 = to_C.chain("x2")
    to_added_2 = restrict(to_C.chain("x2"), term_early_at={"x3"})

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(added_1, to_added_1)
    corr.add(added_2, to_added_2)
    corr.add(x0, to_added_1.chain("x0"))
    corr.add(x1, to_added_2.chain("x1"))

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])
    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)


def test_corr_injective_unused_direct():
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    second = Index(xs, I[1], name="x1")
    C = Add(first, second, name="C")

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    x0_1 = C_t.make_descendant(name="x0_1", cond_sampler=UncondSampler())
    x0_2 = C_t.make_descendant(name="x0_2", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(x0_1, to_C.chain("x0"))
    corr.add(x0_2, restrict(to_C.chain("x0"), term_early_at="x1"))

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])
    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)


def test_corr_tree_structure_same_paths_to_inputs():
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    C = Add(first, name="C")

    # Backwards!
    x0 = InterpNode(UncondSampler(), "x0")
    C_t = x0.make_descendant(name="C", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(x0, to_C.chain("x0"))

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])
    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)


def test_corr_injective_separate_paths():
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    A = Add(first, name="A")
    B = Einsum.scalar_mul(first, 2, name="B")
    C = Add(A, B, name="C")

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    x0_A = C_t.make_descendant(name="x0_A", cond_sampler=UncondSampler())
    x0_B = C_t.make_descendant(name="x0_B", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(x0_A, to_C.chain("A").chain("x0"))
    corr.add(x0_B, to_C.chain("B").chain("x0"))

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])

    ex = Experiment(C, dataset, corr)
    ex.scrub(1, treeify=True, check=True)


def test_corr_tree_structure_respects_paths():
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    A = Add(first, name="A")
    B = Einsum.scalar_mul(first, 2, name="B")
    C = Add(A, B, name="C")

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    A = C_t.make_descendant(name="A", cond_sampler=UncondSampler())
    x0 = A.make_descendant(name="x0", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A, to_C.chain("A"))
    corr.add(x0, to_C.chain("B").chain("x0"))  # not via A!

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])
    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)


def test_corr_tree_structure_only_if():
    # Check that an interp node's paths are extensions of the other's *only if* it is a child of the other
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    A = Einsum.scalar_mul(first, 2, name="A")
    B = Einsum.scalar_mul(A, 3, name="B")
    C = Einsum.scalar_mul(B, 5, name="C")

    # Construct interpretation
    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    A_t = C_t.make_descendant(name="At", cond_sampler=UncondSampler())
    B_t = C_t.make_descendant(name="Bt", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    to_B = to_C.chain("B")
    corr.add(B_t, to_B)
    corr.add(A_t, to_B.chain("A"))

    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])
    with pytest.raises(IllegalCorrespondenceError):
        run_experiment(C, dataset, corr)


# === Unit tests ===


# filter out warnings about check = False
@pytest.mark.filterwarnings("ignore:You're")
def test_treeify():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    A = Einsum.scalar_mul(first, 2, name="A")

    # unused
    dataset = IntDataset.of_shape((5, 2), lambda x: 2 * x[0] + 3 * x[1])

    # Case: child subsumes parent entirely
    C = Einsum.scalar_mul(A, 3, name="C")
    C_t = InterpNode(name="Ct", cond_sampler=ExactSampler())
    out = corr_root_matcher
    to_A = out.chain("A")
    corr = Correspondence()
    corr.add(C_t, out)
    corr.add(C_t.make_descendant(name="A", cond_sampler=ExactSampler()), to_A)
    ex = Experiment(C, dataset, corr, check=False)
    ref_ds = ex.make_ref_ds(1)
    ex.sample(ref_ds)
    circuit = ex.treeified()
    circuit.print()
    assert num_input_tags(circuit) == 1

    # Case: only some children are mapped
    B = Einsum.scalar_mul(second, 3, name="B")
    C = Add(A, B, name="C")
    ex = Experiment(C, dataset, corr)
    ref_ds = ex.make_ref_ds(1)
    ex.sample(ref_ds)
    circuit = ex.treeified()
    assert num_input_tags(circuit) == 2

    # Case: multiple children are mapped
    to_B = out.chain("B")
    corr.add(C_t.make_descendant(name="B", cond_sampler=ExactSampler()), to_B)
    ex = Experiment(C, dataset, corr)
    ref_ds = ex.make_ref_ds(1)
    ex.sample(ref_ds)
    circuit = ex.treeified()
    assert num_input_tags(circuit) == 2


def test_set_inputs():
    # setup
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    A = Einsum.scalar_mul(first, 2, name="A")
    B = Einsum.scalar_mul(second, 3, name="B")
    C = Add(A, B, name="C")
    dataset = IntDataset.of_shape((5, 2), lambda x: 2 * x[0] + 3 * x[1])

    # Create 2 definitely different datasets of the correct len
    A_sample = dataset[0:1]
    B_sample = IntDataset([Array(A_sample.xs.value + 1, "xs"), A_sample.labels])

    # Construct correspondence
    out = corr_root_matcher
    to_A = out.chain("A")
    corr = Correspondence()
    # other_inputs_sampler will be used for the input to B
    C_t = InterpNode(name="C", cond_sampler=ExactSampler(), other_inputs_sampler=FixedOtherSampler(B_sample))
    corr.add(C_t, out)
    corr.add(C_t.make_descendant(name="A", cond_sampler=FixedOtherSampler(A_sample)), to_A)

    # First treeify, then set inputs
    ex = Experiment(C, dataset, corr, random_seed=1)
    ref_ds = ex.make_ref_ds(1)
    sampled_inputs = ex.sample(ref_ds)

    circuit = ex.treeified()
    assert num_input_tags(circuit) == 2

    # Note this number depends on the samplers returning different datasets for the treeified inputs; we've enforced this by construction
    circuit = ex.wrap_in_var(circuit, ref_ds, DiscreteVar.uniform_probs_and_group(1))
    circuit = ex.replace_inputs(circuit, sampled_inputs)
    assert len(IterativeMatcher("xs").get(circuit)) == 2

    # Set inputs without treeifying
    ex = Experiment(C, dataset, corr, random_seed=1)
    ref_ds = ex.make_ref_ds(1)
    sampled_inputs = ex.sample(ref_ds)
    circuit = ex.wrap_in_var(ex.base_circuit, ref_ds, DiscreteVar.uniform_probs_and_group(1))
    circuit = ex.replace_inputs(circuit, sampled_inputs)
    assert len(IterativeMatcher("xs").get(circuit)) == 2


# === E2E tests ===


def test_identical_no_check():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    A = Einsum.scalar_mul(first, 2.0, name="A")
    B = Einsum.scalar_mul(second, 3.0, name="B")
    C = Add(A, B, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((100, 2), lambda x: 2 * x[0] + 3 * x[1])

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=label_cond)
    A_t = C_t.make_descendant(name="At", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    B_t = C_t.make_descendant(name="Bt", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 1]))

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_C.chain("A"))
    corr.add(B_t, to_C.chain("B"))

    run_experiment(C, dataset, corr, lambda l: l == 0, check=False, num_examples=1)


def test_identical():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    A = Einsum.scalar_mul(first, 2.0, name="A")
    B = Einsum.scalar_mul(second, 3.0, name="B")
    C = Add(A, B, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((100, 2), lambda x: 2 * x[0] + 3 * x[1])

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=label_cond)
    A_t = C_t.make_descendant(name="At", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    B_t = C_t.make_descendant(name="Bt", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 1]))

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_C.chain("A"))
    corr.add(B_t, to_C.chain("B"))

    run_experiment(C, dataset, corr, lambda l: l == 0, num_examples=1)


def test_explicit_inputs_okay():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], "x0")
    second = Index(xs, I[1], "x1")
    C = Add(first, second, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((100, 2), lambda x: x[0] + x[1])

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=label_cond)
    x0_t = C_t.make_descendant(name="x0t", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    x1_t = C_t.make_descendant(name="x1t", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 1]))
    xs_x0_t = x0_t.make_descendant(name="x0_xs", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    xs_x1_t = x1_t.make_descendant(name="x1_xs", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 1]))

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(x0_t, to_C.chain("x0"))
    corr.add(x1_t, to_C.chain("x1"))
    corr.add(xs_x0_t, to_C.chain("x0").chain("xs"))
    corr.add(xs_x1_t, to_C.chain("x1").chain("xs"))

    run_experiment(
        C, dataset, corr, lambda l: l == 0, num_examples=1, treeify=False
    )  # explicit inputs lead to failing treeified checks!


def test_spread_over_residual():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x1")
    second = Index(xs, I[1], name="x2")
    A = Add(first, second, name="A")
    C = Add(A, second, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((1000, 2), lambda x: x[0] + 2 * x[1])

    # Check that just running model fails to find issue
    non_scrub_loss = datum_loss(C, dataset)
    assert non_scrub_loss < 0.01

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=label_cond)
    A_t = C_t.make_descendant(name="A", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    x1 = A_t.make_descendant(name="x1", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    x2 = C_t.make_descendant(name="x2", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 1]))

    to_C = corr_root_matcher
    to_A = to_C.chain("A")

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_A)
    corr.add(x1, to_A.chain("x1"))
    corr.add(x2, restrict(to_C.chain("x2"), term_early_at={"A"}))

    C.print()
    print(C.shape)

    run_experiment(C, dataset, corr, lambda l: l > 1, num_examples=50)


def test_spread_over_two_nodes():
    num_examples = 50

    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    A = Einsum.scalar_mul(first, 2, name="A")
    B = Einsum.scalar_mul(first, 3, name="B")
    C = Add(A, B, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((1000, 2), lambda x: 5 * x[0])

    # Check that just running model fails to find issue
    assert datum_loss(C, dataset) < 0.01

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    A_t = C_t.make_descendant(name="A", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))

    to_C = corr_root_matcher
    to_A = to_C.chain("A")

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_A)

    run_experiment(C, dataset, corr, lambda l: l > 0.1, num_examples=num_examples)


def test_large_deviation():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    A = Add(Einsum.scalar_mul(first, 2), second, name="A")
    B = Einsum.scalar_mul(second, 3, name="B")
    C = Add(A, B, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=label_cond)
    A_t = C_t.make_descendant(name="At", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    B_t = C_t.make_descendant(name="Bt", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 1]))

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_C.chain("A"))
    corr.add(B_t, to_C.chain("B"))

    run_experiment(C, dataset, corr, lambda l: l > 1, num_examples=50)


def test_tiny_deviation():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    A = Add(Einsum.scalar_mul(first, 2), Einsum.scalar_mul(second, 0.001), name="A")
    B = Einsum.scalar_mul(second, 3, name="B")
    C = Add(A, B, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((1000, 2), lambda x: 2 * x[0] + 3 * x[1])

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=label_cond)
    A_t = C_t.make_descendant(name="At", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    B_t = C_t.make_descendant(name="Bt", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 1]))

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_C.chain("A"))
    corr.add(B_t, to_C.chain("B"))

    run_experiment(C, dataset, corr, lambda l: l < 0.01, num_examples=1)


def test_intermediates_cancel():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0, 2.0]), name="xs")
    # TODO: fix indexes
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    third = Index(xs, I[2])
    A = Add(first, third, name="A")
    B = Add.minus(second, third, name="B")
    C = Add(A, B, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((1000, 3), lambda x: x[0] + x[1])

    # Check that just running model fails to find issue
    assert datum_loss(C, dataset) < 0.0001

    # Construct interpretation
    C_t = InterpNode(
        name="Ct",
        cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0] + IntDataset.unwrap(d).inp_tensor[:, 1]),
    )
    A_t = C_t.make_descendant(name="At", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    B_t = C_t.make_descendant(name="Bt", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 1]))

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_C.chain("A"))
    corr.add(B_t, to_C.chain("B"))

    run_experiment(C, dataset, corr, lambda l: l > 1, num_examples=50)


def test_swapped_intermediates():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    A = Add(Einsum.scalar_mul(first, 2), second, name="A")
    B = Add(first, Einsum.scalar_mul(second, 2), name="B")
    C = Add(A, B, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((1000, 2), lambda x: 3 * x[0] + 3 * x[1])

    # Check that just running model fails to find issue TODO sample
    assert datum_loss(C, dataset) == 0

    # Construct interpretation
    C_t = InterpNode(
        name="Ct",
        cond_sampler=FS(
            lambda d: 2 * IntDataset.unwrap(d).inp_tensor[:, 0] + 3 * IntDataset.unwrap(d).inp_tensor[:, 1]
        ),
    )
    A_t = C_t.make_descendant(
        name="At",
        cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0] + 2 * IntDataset.unwrap(d).inp_tensor[:, 1]),
    )
    B_t = C_t.make_descendant(
        name="Bt",
        cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0] + IntDataset.unwrap(d).inp_tensor[:, 1]),
    )

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_C.chain("A"))
    corr.add(B_t, to_C.chain("B"))

    run_experiment(C, dataset, corr, lambda l: l > 1, num_examples=50)


def test_data_correlation():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0], name="x0")
    second = Index(xs, I[1], name="x1")
    C = Add(first, second, name="C")

    # Ground truth
    rands = torch.randint(10, size=(1000, 1))
    dataset = IntDataset({"xs": Array(torch.hstack((rands, 2 * rands)), "xs"), "labels": Array(3 * rands, "labels")})

    # Check that just running model fails to find issue
    assert datum_loss(C, dataset) == 0

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    x0_t = C_t.make_descendant(name="x0", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(x0_t, to_C.chain("x0"))

    run_experiment(C, dataset, corr, lambda l: l > 1, num_examples=50)


# filter out warnings about insufficient data
@pytest.mark.filterwarnings("ignore:not enough")
def test_unspecified_nodes():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    second = Index(xs, I[1])
    A = Einsum.scalar_mul(first, 2, name="A")
    B = Einsum.scalar_mul(second, 3, name="B")
    C = Add(A, B, name="C")

    # Ground truth
    dataset = IntDataset.of_shape((5, 2), lambda x: 2 * x[0] + 3 * x[1])

    # Construct interpretation, omitting the path through B
    C_t = InterpNode(name="Ct", cond_sampler=label_cond)
    A_t = C_t.make_descendant(name="At", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_C.chain("A"))

    # By default the unspecified nodes will be sampled randomly
    # so our scrubbed model will do poorly
    run_experiment(C, dataset, corr, lambda l: l > 1, num_examples=50)

    # But we can specify that by default nodes are not resampled...
    C_t.other_inputs_sampler = ExactSampler()
    # in which case we recover perfect performance
    run_experiment(C, dataset, corr, lambda l: l == 0, num_examples=1)


def test_print():
    # Construct model
    xs = Array(torch.tensor([0.0, 1.0]), name="xs")
    first = Index(xs, I[0])
    A = Einsum.scalar_mul(first, 2, name="A")
    B = Einsum.scalar_mul(first, 3, name="B")
    C = Add(A, B, name="C")

    # Deterministic ds
    generator = torch.Generator()
    generator.manual_seed(11)
    dataset = IntDataset(
        (
            Array(torch.randint(high=10, size=(100, 2), generator=generator), "xs"),
            Array(torch.arange(100, 200), "labels"),
        )
    )

    # Construct interpretation
    C_t = InterpNode(name="Ct", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))
    A_t = C_t.make_descendant(name="A", cond_sampler=FS(lambda d: IntDataset.unwrap(d).inp_tensor[:, 0]))

    to_C = corr_root_matcher
    to_A = to_C.chain("A")

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(A_t, to_A)

    ex = Experiment(C, dataset, corr, random_seed=11)
    scrubbed_ex = ex.scrub(1)

    assert (
        scrubbed_ex.print(PrintOptions(), color_by_data=False, repr=True)
        == """C Add # cond_sampler=FuncSampler(d=IntDatum(xs=[0, 5], label=185), f(d)=0), other_inputs_sampler=UncondSampler(d=IntDatum(xs=[2, 3], label=167))
  A Einsum ,-> ... # cond_sampler=FuncSampler(d=IntDatum(xs=[0, 1], label=153), f(d)=0)
  B Einsum ,-> ..."""
    )

    scrubbed_ex.print()  # check coloring doesn't crash, without checking the result
